from __future__ import annotations

import sys
from logging import DEBUG, INFO
from select import select
from subprocess import PIPE

import plumbum.commands.base
from plumbum.commands.processes import BY_TYPE, ProcessExecutionError, run_proc
from plumbum.lib import read_fd_decode_safely


class Future:
    """Represents a "future result" of a running process. It basically wraps a ``Popen``
    object and the expected exit code, and provides poll(), wait(), returncode, stdout,
    and stderr.
    """

    def __init__(self, proc, expected_retcode, timeout=None):
        self.proc = proc
        self._expected_retcode = expected_retcode
        self._timeout = timeout
        self._returncode = None
        self._stdout = None
        self._stderr = None

    def __repr__(self):
        running = self._returncode if self.ready() else "running"
        return f"<Future {self.proc.argv!r} ({running})>"

    def poll(self):
        """Polls the underlying process for termination; returns ``False`` if still running,
        or ``True`` if terminated"""
        if self.proc.poll() is not None:
            self.wait()
        return self._returncode is not None

    ready = poll

    def wait(self):
        """Waits for the process to terminate; will raise a
        :class:`plumbum.commands.ProcessExecutionError` in case of failure"""
        if self._returncode is not None:
            return
        self._returncode, self._stdout, self._stderr = run_proc(
            self.proc, self._expected_retcode, self._timeout
        )

    @property
    def stdout(self):
        """The process' stdout; accessing this property will wait for the process to finish"""
        self.wait()
        return self._stdout

    @property
    def stderr(self):
        """The process' stderr; accessing this property will wait for the process to finish"""
        self.wait()
        return self._stderr

    @property
    def returncode(self):
        """The process' returncode; accessing this property will wait for the process to finish"""
        self.wait()
        return self._returncode


# ===================================================================================================
# execution modifiers
# ===================================================================================================


class ExecutionModifier:
    __slots__ = ("__weakref__",)

    def __repr__(self):
        """Automatically creates a representation for given subclass with slots.
        Ignore hidden properties."""
        slots = {}
        for cls in self.__class__.__mro__:
            slots_list = getattr(cls, "__slots__", ())
            if isinstance(slots_list, str):
                slots_list = (slots_list,)
            for prop in slots_list:
                if prop[0] != "_":
                    slots[prop] = getattr(self, prop)
        mystrs = (f"{name} = {value}" for name, value in slots.items())
        mystrs_str = ", ".join(mystrs)
        return f"{self.__class__.__name__}({mystrs_str})"

    @classmethod
    def __call__(cls, *args, **kwargs):
        return cls(*args, **kwargs)


class _BG(ExecutionModifier):
    """
    An execution modifier that runs the given command in the background, returning a
    :class:`Future <plumbum.commands.Future>` object. In order to mimic shell syntax, it applies
    when you right-and it with a command. If you wish to expect a different return code
    (other than the normal success indicate by 0), use ``BG(retcode)``. Example::

        future = sleep[5] & BG       # a future expecting an exit code of 0
        future = sleep[5] & BG(7)    # a future expecting an exit code of 7

    .. note::

       When processes run in the **background** (either via ``popen`` or
       :class:`& BG <plumbum.commands.BG>`), their stdout/stderr pipes might fill up,
       causing them to hang. If you know a process produces output, be sure to consume it
       every once in a while, using a monitoring thread/reactor in the background.
       For more info, see `#48 <https://github.com/tomerfiliba/plumbum/issues/48>`_
    """

    __slots__ = ("kargs", "retcode", "timeout")

    def __init__(self, retcode=0, timeout=None, **kargs):
        self.retcode = retcode
        self.kargs = kargs
        self.timeout = timeout

    def __rand__(self, cmd):
        return Future(cmd.popen(**self.kargs), self.retcode, timeout=self.timeout)


class _FG(ExecutionModifier):
    """
    An execution modifier that runs the given command in the foreground, passing it the
    current process' stdin, stdout and stderr. Useful for interactive programs that require
    a TTY. There is no return value.

    In order to mimic shell syntax, it applies when you right-and it with a command.
    If you wish to expect a different return code (other than the normal success indicate by 0),
    use ``FG(retcode)``. Example::

        vim & FG       # run vim in the foreground, expecting an exit code of 0
        vim & FG(7)    # run vim in the foreground, expecting an exit code of 7
    """

    __slots__ = ("retcode", "timeout")

    def __init__(self, retcode=0, timeout=None):
        self.retcode = retcode
        self.timeout = timeout

    def __rand__(self, cmd):
        cmd(
            retcode=self.retcode,
            stdin=None,
            stdout=None,
            stderr=None,
            timeout=self.timeout,
        )


class _TEE(ExecutionModifier):
    """Run a command, dumping its stdout/stderr to the current process's stdout
    and stderr, but ALSO return them.  Useful for interactive programs that
    expect a TTY but also have valuable output.

    Use as:

        ls["-l"] & TEE

    Returns a tuple of (return code, stdout, stderr), just like ``run()``.
    """

    __slots__ = ("buffered", "retcode", "timeout")

    def __init__(self, retcode=0, buffered=True, timeout=None):
        """`retcode` is the return code to expect to mean "success".  Set
        `buffered` to False to disable line-buffering the output, which may
        cause stdout and stderr to become more entangled than usual.
        """
        self.retcode = retcode
        self.buffered = buffered
        self.timeout = timeout

    def __rand__(self, cmd):
        with cmd.bgrun(
            retcode=self.retcode,
            stdin=None,
            stdout=PIPE,
            stderr=PIPE,
            timeout=self.timeout,
        ) as p:
            outbuf = []
            errbuf = []
            out = p.stdout
            err = p.stderr
            buffers = {out: outbuf, err: errbuf}
            tee_to = {out: sys.stdout, err: sys.stderr}
            done = False
            while not done:
                # After the process exits, we have to do one more
                # round of reading in order to drain any data in the
                # pipe buffer. Thus, we check poll() here,
                # unconditionally enter the read loop, and only then
                # break out of the outer loop if the process has
                # exited.
                done = p.poll() is not None

                # We continue this loop until we've done a full
                # `select()` call without collecting any input. This
                # ensures that our final pass -- after process exit --
                # actually drains the pipe buffers, even if it takes
                # multiple calls to read().
                progress = True
                while progress:
                    progress = False
                    ready, _, _ = select((out, err), (), ())
                    for fd in ready:
                        buf = buffers[fd]
                        data, text = read_fd_decode_safely(fd, 4096)
                        if not data:  # eof
                            continue
                        progress = True

                        # Python conveniently line-buffers stdout and stderr for
                        # us, so all we need to do is write to them

                        # This will automatically add up to three bytes if it cannot be decoded
                        tee_to[fd].write(text)

                        # And then "unbuffered" is just flushing after each write
                        if not self.buffered:
                            tee_to[fd].flush()

                        buf.append(data)

            p.wait()  # To get return code in p
            stdout = "".join([x.decode("utf-8") for x in outbuf])
            stderr = "".join([x.decode("utf-8") for x in errbuf])
            return p.returncode, stdout, stderr


class _TF(ExecutionModifier):
    """
    An execution modifier that runs the given command, but returns True/False depending on the retcode.
    This returns True if the expected exit code is returned, and false if it is not.
    This is useful for checking true/false bash commands.

    If you wish to expect a different return code (other than the normal success indicate by 0),
    use ``TF(retcode)``. If you want to run the process in the foreground, then use
    ``TF(FG=True)``.

    Example::

        local['touch']['/root/test'] & TF * Returns False, since this cannot be touched
        local['touch']['/root/test'] & TF(1) # Returns True
        local['touch']['/root/test'] & TF(FG=True) * Returns False, will show error message
    """

    __slots__ = ("FG", "retcode", "timeout")

    def __init__(
        self,
        retcode=0,
        FG=False,  # pylint: disable=redefined-outer-name
        timeout=None,
    ):
        """`retcode` is the return code to expect to mean "success".  Set
        `FG` to True to run in the foreground.
        """
        self.retcode = retcode
        self.FG = FG
        self.timeout = timeout

    @classmethod
    def __call__(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def __rand__(self, cmd):
        try:
            if self.FG:
                cmd(
                    retcode=self.retcode,
                    stdin=None,
                    stdout=None,
                    stderr=None,
                    timeout=self.timeout,
                )
            else:
                cmd(retcode=self.retcode, timeout=self.timeout)
            return True
        except ProcessExecutionError:
            return False


class _RETCODE(ExecutionModifier):
    """
    An execution modifier that runs the given command, causing it to run and return the retcode.
    This is useful for working with bash commands that have important retcodes but not very
    useful output.

    If you want to run the process in the foreground, then use ``RETCODE(FG=True)``.

    Example::

        local['touch']['/root/test'] & RETCODE # Returns 1, since this cannot be touched
        local['touch']['/root/test'] & RETCODE(FG=True) * Returns 1, will show error message
    """

    __slots__ = ("foreground", "timeout")

    def __init__(
        self,
        FG=False,  # pylint: disable=redefined-outer-name
        timeout=None,
    ):
        """`FG` to True to run in the foreground."""
        self.foreground = FG
        self.timeout = timeout

    @classmethod
    def __call__(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def __rand__(self, cmd):
        if self.foreground:
            result = cmd.run(
                retcode=None, stdin=None, stdout=None, stderr=None, timeout=self.timeout
            )
            return result[0]

        return cmd.run(retcode=None, timeout=self.timeout)[0]


class _NOHUP(ExecutionModifier):
    """
    An execution modifier that runs the given command in the background, disconnected
    from the current process, returning a
    standard popen object. It will keep running even if you close the current process.
    In order to slightly mimic shell syntax, it applies
    when you right-and it with a command. If you wish to use a different working directory
    or different stdout, stderr, you can use named arguments. The default is ``NOHUP(
    cwd=local.cwd, stdout='nohup.out', stderr=None)``. If stderr is None, stderr will be
    sent to stdout. Use ``os.devnull`` for null output. Will respect redirected output.
    Example::

        sleep[5] & NOHUP                       # Outputs to nohup.out
        sleep[5] & NOHUP(stdout=os.devnull)    # No output

    The equivalent bash command would be

    .. code-block:: bash

        nohup sleep 5 &

    """

    __slots__ = ("append", "cwd", "stderr", "stdout")

    def __init__(self, cwd=".", stdout="nohup.out", stderr=None, append=True):
        """Set ``cwd``, ``stdout``, or ``stderr``.
        Runs as a forked process. You can set ``append=False``, too.
        """
        self.cwd = cwd
        self.stdout = stdout
        self.stderr = stderr
        self.append = append

    def __rand__(self, cmd):
        if isinstance(cmd, plumbum.commands.base.StdoutRedirection):
            stdout = cmd.file
            append = False
            cmd = cmd.cmd
        elif isinstance(cmd, plumbum.commands.base.AppendingStdoutRedirection):
            stdout = cmd.file
            append = True
            cmd = cmd.cmd
        else:
            stdout = self.stdout
            append = self.append
        return cmd.nohup(self.cwd, stdout, self.stderr, append)


class LogPipe:
    def __init__(self, line_timeout, kw, levels, prefix, log):
        self.line_timeout = line_timeout
        self.kw = kw
        self.levels = levels
        self.prefix = prefix
        self.log = log

    def __rand__(self, cmd):
        popen = cmd if hasattr(cmd, "iter_lines") else cmd.popen()
        for typ, lines in popen.iter_lines(
            line_timeout=self.line_timeout, mode=BY_TYPE, **self.kw
        ):
            if not lines:
                continue
            level = self.levels[typ]
            for line in lines.splitlines():
                if self.prefix:
                    line = f"{self.prefix}: {line}"  # noqa: PLW2901
                self.log(level, line)
        return popen.returncode


class PipeToLoggerMixin:
    """
    This mixin allows piping plumbum commands' output into a logger.
    The logger must implement a ``log(level, msg)`` method, as in ``logging.Logger``

    Example::

        class MyLogger(logging.Logger, PipeToLoggerMixin):
            pass

        logger = MyLogger("example.app")

    Here we send the output of an install.sh script into our log::

        local['./install.sh'] & logger

    We can choose the log-level for each stream::

        local['./install.sh'] & logger.pipe(out_level=logging.DEBUG, err_level=logging.DEBUG)

    Or use a convenience method for it::

        local['./install.sh'] & logger.pipe_debug()

    A prefix can be added to each line::

        local['./install.sh'] & logger.pipe(prefix="install.sh: ")

    If the command fails, an exception is raised as usual. This can be modified::

        local['install.sh'] & logger.pipe_debug(retcode=None)

    An exception is also raised if too much time (``DEFAULT_LINE_TIMEOUT``) passed between lines in the stream,
    This can also be modified::

        local['install.sh'] & logger.pipe(line_timeout=10)

    If we happen to use logbook::

        class MyLogger(logbook.Logger, PipeToLoggerMixin):
            from logbook import DEBUG, INFO  # hook up with logbook's levels

    """

    DEFAULT_LINE_TIMEOUT = 10 * 60
    DEFAULT_STDOUT = "INFO"
    DEFAULT_STDERR = "DEBUG"
    INFO = INFO
    DEBUG = DEBUG

    def pipe(
        self, out_level=None, err_level=None, prefix=None, line_timeout=None, **kw
    ):
        """
        Pipe a command's stdout and stderr lines into this logger.

        :param out_level: the log level for lines coming from stdout
        :param err_level: the log level for lines coming from stderr

        Optionally use `prefix` for each line.
        """

        levels = {
            1: getattr(self, self.DEFAULT_STDOUT),
            2: getattr(self, self.DEFAULT_STDERR),
        }

        if line_timeout is None:
            line_timeout = self.DEFAULT_LINE_TIMEOUT

        if out_level is not None:
            levels[1] = out_level

        if err_level is not None:
            levels[2] = err_level

        return LogPipe(line_timeout, kw, levels, prefix, self.log)

    def pipe_info(self, prefix=None, **kw):
        """
        Pipe a command's stdout and stderr lines into this logger (both at level INFO)
        """
        return self.pipe(self.INFO, self.INFO, prefix=prefix, **kw)

    def pipe_debug(self, prefix=None, **kw):
        """
        Pipe a command's stdout and stderr lines into this logger (both at level DEBUG)
        """
        return self.pipe(self.DEBUG, self.DEBUG, prefix=prefix, **kw)

    def __rand__(self, cmd):
        """
        Pipe a command's stdout and stderr lines into this logger.
        Log levels for each stream are determined by ``DEFAULT_STDOUT`` and ``DEFAULT_STDERR``.
        """
        return cmd & self.pipe(
            getattr(self, self.DEFAULT_STDOUT), getattr(self, self.DEFAULT_STDERR)
        )


BG = _BG()
FG = _FG()
NOHUP = _NOHUP()
RETCODE = _RETCODE()
TEE = _TEE()
TF = _TF()



from __future__ import annotations

import os
import pickle
import signal
import sys
import time
from pathlib import Path

import pytest

import plumbum
from plumbum import (
    BG,
    ERROUT,
    FG,
    RETCODE,
    TEE,
    TF,
    CommandNotFound,
    LocalPath,
    ProcessExecutionError,
    ProcessLineTimedOut,
    ProcessTimedOut,
    local,
)
from plumbum._testtools import skip_on_windows, skip_without_chown, skip_without_tty
from plumbum.fs.atomic import AtomicCounterFile, AtomicFile, PidFile
from plumbum.lib import IS_WIN32
from plumbum.machines.local import LocalCommand, PlumbumLocalPopen
from plumbum.path import RelativePath

# This is a string since we are testing local paths
SDIR = os.path.dirname(os.path.abspath(__file__))


class TestLocalPopen:
    def test_contextmanager(self):
        command = ["dir"] if IS_WIN32 else ["ls"]
        with PlumbumLocalPopen(command):
            pass


class TestLocalPath:
    longpath = local.path("/some/long/path/to/file.txt")

    def test_name(self):
        name = self.longpath.name
        assert isinstance(name, str)
        assert str(name) == "file.txt"

    def test_dirname(self):
        name = self.longpath.dirname
        assert isinstance(name, LocalPath)
        assert (
            str(name).replace("\\", "/").lstrip("C:").lstrip("D:")
            == "/some/long/path/to"
        )

    def test_uri(self):
        if IS_WIN32:
            pth = self.longpath.as_uri()
            assert pth.startswith("file:///")
            assert pth.endswith(":/some/long/path/to/file.txt")
        else:
            assert self.longpath.as_uri() == "file:///some/long/path/to/file.txt"

    def test_pickle(self):
        path1 = local.path(".")
        path2 = local.path("~")
        assert pickle.loads(pickle.dumps(self.longpath)) == self.longpath
        assert pickle.loads(pickle.dumps(path1)) == path1
        assert pickle.loads(pickle.dumps(path2)) == path2

    def test_empty(self):
        with pytest.raises(TypeError):
            LocalPath()
        assert local.path() == local.path(".")

    @skip_without_chown
    def test_chown(self):
        with local.tempdir() as dir:
            p = dir / "foo.txt"
            p.write(b"hello")
            assert p.uid == os.getuid()
            assert p.gid == os.getgid()
            p.chown(p.uid.name)
            assert p.uid == os.getuid()

    def test_split(self):
        p = local.path("/var/log/messages")
        assert p.split() == ["var", "log", "messages"]

    def test_suffix(self):
        # This picks up the drive letter differently if not constructed here
        p1 = local.path("/some/long/path/to/file.txt")
        p2 = local.path("file.tar.gz")
        assert p1.suffix == ".txt"
        assert p1.suffixes == [".txt"]
        assert p2.suffix == ".gz"
        assert p2.suffixes == [".tar", ".gz"]
        assert p1.with_suffix(".tar.gz") == local.path("/some/long/path/to/file.tar.gz")
        assert p2.with_suffix(".other") == local.path("file.tar.other")
        assert p2.with_suffix(".other", 2) == local.path("file.other")
        assert p2.with_suffix(".other", 0) == local.path("file.tar.gz.other")
        assert p2.with_suffix(".other", None) == local.path("file.other")

        with pytest.raises(ValueError):
            p1.with_suffix("nodot")

    def test_newname(self):
        # This picks up the drive letter differently if not constructed here
        p1 = local.path("/some/long/path/to/file.txt")
        p2 = local.path("file.tar.gz")
        assert p1.with_name("something.tar") == local.path(
            "/some/long/path/to/something.tar"
        )
        assert p2.with_name("something.tar") == local.path("something.tar")

    def test_relative_to(self):
        p = local.path("/var/log/messages")
        assert p.relative_to("/var/log/messages") == RelativePath([])
        assert p.relative_to("/var/") == RelativePath(["log", "messages"])
        assert p.relative_to("/") == RelativePath(["var", "log", "messages"])
        assert p.relative_to("/var/tmp") == RelativePath(["..", "log", "messages"])
        assert p.relative_to("/opt") == RelativePath(["..", "var", "log", "messages"])
        assert p.relative_to("/opt/lib") == RelativePath(
            ["..", "..", "var", "log", "messages"]
        )
        for src in [
            local.path("/var/log/messages"),
            local.path("/var"),
            local.path("/opt/lib"),
        ]:
            delta = p.relative_to(src)
            assert src + delta == p

    def test_read_write(self):
        with local.tempdir() as dir:
            f = dir / "test.txt"
            text = b"hello world\xd7\xa9\xd7\x9c\xd7\x95\xd7\x9d".decode("utf8")
            f.write(text, "utf8")
            text2 = f.read("utf8")
            assert text == text2

    def test_parts(self):
        parts = self.longpath.parts
        if IS_WIN32:
            assert parts[1:] == ("some", "long", "path", "to", "file.txt")
            assert ":" in parts[0]
        else:
            assert parts == ("/", "some", "long", "path", "to", "file.txt")

    @pytest.mark.usefixtures("testdir")
    def test_iterdir(self):
        cwd = local.path(".")
        files = list(cwd.iterdir())
        assert cwd / "test_local.py" in files
        assert cwd / "test_remote.py" in files
        assert cwd["test_local.py"] in files
        assert cwd["test_remote.py"] in files

    def test_stem(self):
        assert self.longpath.stem == "file"
        p = local.path("/some/directory")
        assert p.stem == "directory"

    def test_root_drive(self):
        pathlib = pytest.importorskip("pathlib")
        pl_path = pathlib.Path("/some/long/path/to/file.txt").absolute()
        assert self.longpath.root == pl_path.root
        assert self.longpath.drive == pl_path.drive

        p_path = local.cwd / "somefile.txt"
        pl_path = pathlib.Path("somefile.txt").absolute()
        assert p_path.root == pl_path.root
        assert p_path.drive == pl_path.drive

    def test_compare_pathlib(self):
        pathlib = pytest.importorskip("pathlib")

        def filename_compare(name):
            p = local.path(str(name))
            pl = pathlib.Path(str(name)).absolute()
            assert str(p) == str(pl)
            assert p.parts == pl.parts
            assert p.exists() == pl.exists()
            assert p.is_symlink() == pl.is_symlink()
            assert p.as_uri() == pl.as_uri()
            assert str(p.with_suffix(".this")) == str(pl.with_suffix(".this"))
            assert p.name == pl.name
            assert str(p.parent) == str(pl.parent)
            assert list(map(str, p.parents)) == list(map(str, pl.parents))

        filename_compare("/some/long/path/to/file.txt")
        filename_compare(local.cwd / "somefile.txt")
        filename_compare("/some/long/path/")
        filename_compare("/some/long/path")
        filename_compare(__file__)

    def test_suffix_expected(self):
        assert self.longpath.preferred_suffix(".tar") == self.longpath
        assert (local.cwd / "this").preferred_suffix(".txt") == local.cwd / "this.txt"

    def test_touch(self):
        with local.tempdir() as tmp:
            one = tmp / "one"
            assert not one.is_file()
            one.touch()
            assert one.is_file()
            one.delete()
            assert not one.is_file()

    def test_copy_override(self):
        """Edit this when override behavior is added"""
        with local.tempdir() as tmp:
            one = tmp / "one"
            one.touch()
            two = tmp / "two"
            assert one.is_file()
            assert not two.is_file()
            one.copy(two)
            assert one.is_file()
            assert two.is_file()

    def test_copy_nonexistant_dir(self):
        with local.tempdir() as tmp:
            one = tmp / "one"
            one.write(b"lala")
            two = tmp / "two" / "one"
            three = tmp / "three" / "two" / "one"

            one.copy(two)
            assert one.read() == two.read()
            one.copy(three)
            assert one.read() == three.read()

    def test_unlink(self):
        with local.tempdir() as tmp:
            one = tmp / "one"
            one.touch()
            assert one.exists()
            one.unlink()
            assert not one.exists()

    def test_unhashable(self):
        with pytest.raises(TypeError):
            hash(local.cwd)

    def test_getpath(self):
        assert local.cwd.getpath() == local.path(".")

    def test_path_dir(self):
        assert local.path(__file__).dirname == SDIR

    def test_mkdir(self):
        # (identical to test_remote.TestRemotePath.test_mkdir)
        with local.tempdir() as tmp:
            tmp["a"].mkdir(exist_ok=False, parents=False)
            assert tmp["a"].exists()
            assert tmp["a"].is_dir()
            tmp["a"].mkdir(exist_ok=True, parents=False)
            tmp["a"].mkdir(exist_ok=True, parents=True)
            with pytest.raises(OSError):
                tmp["a"].mkdir(exist_ok=False, parents=False)
            with pytest.raises(OSError):
                tmp["a"].mkdir(exist_ok=False, parents=True)
            tmp["b"]["bb"].mkdir(exist_ok=False, parents=True)
            assert tmp["b"]["bb"].exists()
            assert tmp["b"]["bb"].is_dir()
        assert not tmp.exists()

    def test_mkdir_mode(self):
        # (identical to test_remote.TestRemotePath.test_mkdir_mode)
        with local.tempdir() as tmp:
            # just verify that mode argument works the same way it does for
            # Python's own os.mkdir, which takes into account the umask
            # (different from shell mkdir mode argument!); umask on my
            # system is 022 by default, so 033 is ok for testing this
            try:
                (tmp / "pb_333").mkdir(exist_ok=False, parents=False, mode=0o333)
                local.python(
                    "-c",
                    "import os; os.mkdir({}, 0o333)".format(repr(str(tmp / "py_333"))),
                )
                pb_final_mode = oct((tmp / "pb_333").stat().st_mode)
                py_final_mode = oct((tmp / "py_333").stat().st_mode)
                assert pb_final_mode == py_final_mode
            finally:
                # we have to revert this so the tempdir deletion works
                if (tmp / "pb_333").exists():
                    (tmp / "pb_333").chmod(0o777)
                if (tmp / "py_333").exists():
                    (tmp / "py_333").chmod(0o777)
        assert not tmp.exists()

    def test_str_getitem(self):
        with local.tempdir() as tmp:
            assert str(tmp) == str(tmp[:])
            assert str(tmp)[0] == str(tmp[0])

    def test_fspath(self):
        with local.tempdir() as tmp:
            assert tmp.__fspath__() == str(tmp)


@pytest.mark.usefixtures("testdir")
class TestLocalMachine:
    def test_getattr(self):
        pb = plumbum
        assert getattr(pb.cmd, "does_not_exist", 1) == 1
        ls_cmd1 = pb.cmd.non_exist1N9 if hasattr(pb.cmd, "non_exist1N9") else pb.cmd.ls
        ls_cmd2 = getattr(pb.cmd, "non_exist1N9", pb.cmd.ls)
        assert str(ls_cmd1) == str(local["ls"])
        assert str(ls_cmd2) == str(local["ls"])

    # TODO: This probably fails because of odd ls behavior
    @skip_on_windows
    def test_imports(self):
        from plumbum.cmd import ls

        assert "test_local.py" in local["ls"]().splitlines()
        assert "test_local.py" in ls().splitlines()

        with pytest.raises(CommandNotFound):
            local["non_exist1N9"]()

        with pytest.raises(ImportError):
            from plumbum.cmd import non_exist1N9  # noqa: F401

    def test_pathlib(self):
        ls_path = Path(local.which("ls"))
        assert "test_local.py" in local[ls_path]().splitlines()

    def test_get(self):
        assert str(local["ls"]) == str(local.get("ls"))
        assert str(local["ls"]) == str(local.get("non_exist1N9", "ls"))

        with pytest.raises(CommandNotFound):
            local.get("non_exist1N9")
        with pytest.raises(CommandNotFound):
            local.get("non_exist1N9", "non_exist1N8")
        with pytest.raises(CommandNotFound):
            local.get("non_exist1N9", "/tmp/non_exist1N8")

    def test_shadowed_by_dir(self):
        real_ls = local["ls"]
        with local.tempdir() as tdir, local.cwd(tdir):
            ls_dir = tdir / "ls"
            ls_dir.mkdir()
            fake_ls = local["ls"]
            assert fake_ls.executable == real_ls.executable

            local.env.path.insert(0, tdir)
            fake_ls = local["ls"]
            del local.env.path[0]
            assert fake_ls.executable == real_ls.executable

    def test_repr_command(self):
        assert "BG" in repr(BG)
        assert "FG" in repr(FG)

    @skip_on_windows
    def test_cwd(self):
        from plumbum.cmd import ls

        assert local.cwd == os.getcwd()
        assert "machines" not in ls().splitlines()

        with local.cwd("../plumbum"):
            assert "machines" in ls().splitlines()
        assert "machines" not in ls().splitlines()

        assert "machines" in ls.with_cwd("../plumbum")().splitlines()
        path = local.cmd.pwd.with_cwd("../plumbum")().strip()
        with local.cwd("/"):
            assert "machines" not in ls().splitlines()
            assert "machines" in ls.with_cwd(path)().splitlines()

        with pytest.raises(OSError):
            local.cwd.chdir("../non_exist1N9")

    @skip_on_windows
    def test_mixing_chdir(self):
        assert local.cwd == os.getcwd()
        os.chdir("../plumbum")
        assert local.cwd == os.getcwd()
        os.chdir("../tests")
        assert local.cwd == os.getcwd()

    def test_contains(self):
        assert "plumbum" in local.cwd / ".."
        assert "non_exist1N91" not in local.cwd / ".."

    @skip_on_windows
    def test_path(self):
        assert not (local.cwd / "../non_exist1N9").exists()
        assert (local.cwd / ".." / "plumbum").is_dir()
        # traversal
        found = False
        for fn in local.cwd / ".." / "plumbum":
            if fn.name == "__init__.py":
                assert fn.is_file()
                found = True
        assert found
        # glob'ing
        found = False
        for fn in local.cwd / ".." // "*/*.rst":
            if fn.name == "index.rst":
                found = True
        assert found
        for fn in local.cwd / ".." // ("*/*.rst", "*./*.html"):
            if fn.name == "index.rst":
                found = True
        assert found

    @skip_on_windows
    def test_glob_spaces(self):
        fileloc = local.cwd / "file with space.txt"
        assert fileloc.exists()

        assert local.cwd // "*space.txt"
        assert local.cwd // "file with*"

    @skip_on_windows
    def test_env(self):
        assert "PATH" in local.env
        assert "FOOBAR72" not in local.env
        with pytest.raises(ProcessExecutionError):
            local.python("-c", "import os;os.environ['FOOBAR72']")
        local.env["FOOBAR72"] = "spAm"
        assert local.python(
            "-c", "import os; print(os.environ['FOOBAR72'])"
        ).splitlines() == ["spAm"]

        with local.env(FOOBAR73=1889):
            assert local.python(
                "-c", "import os; print(os.environ['FOOBAR73'])"
            ).splitlines() == ["1889"]
            with local.env(FOOBAR73=1778):
                assert local.python(
                    "-c", "import os; print(os.environ['FOOBAR73'])"
                ).splitlines() == ["1778"]
            assert local.python(
                "-c", "import os; print(os.environ['FOOBAR73'])"
            ).splitlines() == ["1889"]
        with pytest.raises(ProcessExecutionError):
            local.python("-c", "import os; os.environ['FOOBAR73']")

        # path manipulation
        with pytest.raises(CommandNotFound):
            local.which("dummy-executable")
        with local.env():
            local.env.path.insert(0, local.cwd / "not-in-path")
            p = local.which("dummy-executable")
            assert p == local.cwd / "not-in-path" / "dummy-executable"

    def test_local(self):
        assert "plumbum" in str(local.cwd)
        assert "PATH" in local.env.getdict()
        assert local.path("foo") == os.path.join(os.getcwd(), "foo")
        local.which("ls")
        local["ls"]
        assert local.python("-c", "print('hi there')").splitlines() == ["hi there"]

    @skip_on_windows
    def test_piping(self):
        from plumbum.cmd import grep, ls

        chain = ls | grep["\\.py"]
        assert "test_local.py" in chain().splitlines()

        chain = ls["-a"] | grep["test"] | grep["local"]
        assert "test_local.py" in chain().splitlines()

    @skip_on_windows
    def test_redirection(self):
        from plumbum.cmd import cat, grep, ls, rm

        chain = (ls | grep["\\.py"]) > "tmp.txt"
        chain()

        chain2 = (cat < "tmp.txt") | grep["local"]
        assert "test_local.py" in chain2().splitlines()
        rm("tmp.txt")

        chain3 = (
            cat << "this is the\nworld of helloness and\nspam bar and eggs"
        ) | grep["hello"]
        assert "world of helloness and" in chain3().splitlines()

        rc, _, err = (grep["-Zq5"] >= "tmp2.txt").run(["-Zq5"], retcode=None)
        assert rc == 2
        assert not err
        assert "usage" in (cat < "tmp2.txt")().lower()
        rm("tmp2.txt")

        rc, out, _ = (grep["-Zq5"] >= ERROUT).run(["-Zq5"], retcode=None)
        assert rc == 2
        assert "usage" in out.lower()

    @skip_on_windows
    def test_popen(self):
        from plumbum.cmd import ls

        p = ls.popen(["-a"])
        out, _ = p.communicate()
        assert p.returncode == 0
        assert "test_local.py" in out.decode(local.encoding).splitlines()

    def test_run(self):
        from plumbum.cmd import grep, ls

        rc, _out, _err = (ls | grep["non_exist1N9"]).run(retcode=1)
        assert rc == 1

    def test_timeout(self):
        from plumbum.cmd import sleep

        with pytest.raises(ProcessTimedOut):
            sleep(3, timeout=1)

    @skip_on_windows
    def test_pipe_stderr(self, capfd):
        from plumbum.cmd import cat, head

        cat["/dev/urndom"] & FG(1)
        assert "urndom" in capfd.readouterr()[1]

        assert capfd.readouterr()[1] == ""

        (cat["/dev/urndom"] | head["-c", "10"]) & FG(retcode=1)
        assert "urndom" in capfd.readouterr()[1]

    @skip_on_windows
    def test_fair_error_attribution(self):
        # use LocalCommand directly for predictable argv
        false = LocalCommand("false")
        true = LocalCommand("true")
        with pytest.raises(ProcessExecutionError) as e:
            (false | true) & FG
        assert e.value.argv == ["false"]

    @skip_on_windows
    def test_iter_lines_timeout(self):
        from plumbum.cmd import bash

        cmd = bash["-ce", "for ((i=0;1==1;i++)); do echo $i; sleep .3; done"]
        with pytest.raises(ProcessTimedOut):  # noqa: PT012
            for i, (out, err) in enumerate(cmd.popen().iter_lines(timeout=1)):
                assert not err
                assert out
                print(i, "out:", out)
        assert i in (2, 3)  # Mac is a bit flakey

    @skip_on_windows
    def test_iter_lines_buffer_size(self):
        from plumbum.cmd import bash

        cmd = bash["-ce", "for ((i=0;i<100;i++)); do echo $i; done; false"]
        with pytest.raises(ProcessExecutionError) as e:
            for _ in cmd.popen().iter_lines(timeout=1, buffer_size=5):
                pass
        assert e.value.stdout == "\n".join(map(str, range(95, 100))) + "\n"

    @skip_on_windows
    def test_iter_lines_timeout_by_type(self):
        from plumbum.cmd import bash
        from plumbum.commands.processes import BY_TYPE

        cmd = bash[
            "-ce", "for ((i=0;1==1;i++)); do echo $i; sleep .3; echo $i 1>&2; done"
        ]
        types = {1: "out:", 2: "err:"}
        counts = {1: 0, 2: 0}
        with pytest.raises(ProcessTimedOut):  # noqa: PT012
            # Order is important on mac
            for typ, line in cmd.popen().iter_lines(timeout=1, mode=BY_TYPE):
                counts[typ] += 1
                print(types[typ], line)
        assert counts[1] in (3, 4)  # Mac is a bit flakey
        assert counts[2] in (2, 3)  # Mac is a bit flakey

    @skip_on_windows
    def test_iter_lines_error(self):
        from plumbum.cmd import ls

        with pytest.raises(ProcessExecutionError) as err:  # noqa: PT012
            for i, _lines in enumerate(ls["--bla"].popen()):  # noqa: B007
                pass
            assert i == 1
        assert (
            "ls: unrecognized option" in err.value.stderr
            and "--bla" in err.value.stderr
        ) or "ls: illegal option -- -" in err.value.stderr

    @skip_on_windows
    def test_iter_lines_line_timeout(self):
        from plumbum.cmd import bash

        cmd = bash["-ce", "for ((i=0;1==1;i++)); do echo $i; sleep $i; done"]

        with pytest.raises(ProcessLineTimedOut):  # noqa: PT012
            # Order is important on mac
            for i, (out, err) in enumerate(cmd.popen().iter_lines(line_timeout=0.2)):
                print(i, "out:", out)
                print(i, "err:", err)
        assert i == 1

    @skip_on_windows
    def test_modifiers(self):
        from plumbum.cmd import cat, grep, ls

        f = (ls["-a"] | grep["\\.py"]) & BG
        f.wait()
        assert "test_local.py" in f.stdout.splitlines()

        command = ls["-a"] | grep["local"]
        command_false = ls["-a"] | grep["not_a_file_here"]
        command_false_2 = command_false | cat
        command & FG
        assert command & TF
        assert not (command_false & TF)
        assert not (command_false_2 & TF)
        assert command & RETCODE == 0
        assert command_false & RETCODE == 1
        assert command_false_2 & RETCODE == 1
        assert (command & TEE)[0] == 0
        assert (command_false & TEE(retcode=None))[0] == 1
        assert (command_false_2 & TEE(retcode=None))[0] == 1

    @skip_on_windows
    def test_tee_modifier(self, capfd):
        from plumbum.cmd import echo

        result = echo["This is fun"] & TEE
        assert result[1] == "This is fun\n"
        assert capfd.readouterr()[0] == "This is fun\n"

    @skip_on_windows
    def test_tee_race(self, capfd):
        from plumbum.cmd import seq

        EXPECT = "".join(f"{i}\n" for i in range(1, 5001))
        for _ in range(5):
            result = seq["1", "5000"] & TEE
            assert result[1] == EXPECT
            assert capfd.readouterr()[0] == EXPECT

    @skip_on_windows
    @pytest.mark.parametrize(
        ("modifier", "expected"),
        [
            (FG, None),
            (TF(FG=True), True),
            (RETCODE(FG=True), 0),
            (TEE, (0, "meow", "")),
        ],
    )
    def test_redirection_stdin_modifiers_fg(self, modifier, expected, capfd):
        "StdinDataRedirection compatible with modifiers which write to stdout"
        from plumbum.cmd import cat

        cmd = cat << "meow"

        assert cmd & modifier == expected
        assert capfd.readouterr() == ("meow", "")

    @skip_on_windows
    def test_logger_pipe(self):
        from plumbum.cmd import bash
        from plumbum.commands.modifiers import PipeToLoggerMixin

        logs = []

        class Logger(PipeToLoggerMixin):
            def log(self, level, line):
                print(level, line)
                logs.append((level, line))

        logger = Logger()

        ret = bash["-ce", "echo aaa"] & logger
        assert logs[-1] == (PipeToLoggerMixin.INFO, "aaa")
        assert ret == 0

        bash["-ce", "echo bbb 1>&2"] & logger
        assert logs[-1] == (PipeToLoggerMixin.DEBUG, "bbb")

        ret = bash["-ce", "echo ccc 1>&2; false"] & logger.pipe(
            prefix="echo", retcode=1, err_level=0
        )
        assert logs[-1] == (0, "echo: ccc")
        assert ret == 1

    @skip_on_windows
    def test_logger_pipe_line_timeout(self):
        from plumbum.cmd import bash
        from plumbum.commands.modifiers import PipeToLoggerMixin

        cmd = bash["-ce", "for ((i=0;i<10;i++)); do echo .$i; sleep .$i; done"]

        class Logger(PipeToLoggerMixin):
            def log(self, level, line):
                print(level, line)
                assert level == 20
                assert float(line) <= 0.6

        logger = Logger()

        with pytest.raises(ProcessLineTimedOut):
            # Order is important on mac
            cmd & logger.pipe(line_timeout=0.45)

    def test_arg_expansion(self):
        from plumbum.cmd import ls

        args = ["-l", "-F"]
        ls(*args)
        ls[args]

    @skip_on_windows
    def test_session(self):
        sh = local.session()
        for _ in range(4):
            _, out, _ = sh.run("ls -a")
            assert "test_local.py" in out.splitlines()

        sh.run("cd ..")
        sh.run("export FOO=17")
        out = sh.run("echo $FOO")[1]
        assert out.splitlines() == ["17"]

    def test_quoting(self):
        ssh = local["ssh"]
        pwd = local["pwd"]

        cmd = ssh[
            "localhost",
            "cd",
            "/usr",
            "&&",
            ssh[
                "localhost", "cd", "/", "&&", ssh["localhost", "cd", "/bin", "&&", pwd]
            ],
        ]
        assert "\"'&&'\"" in " ".join(cmd.formulate(0))

        ls = local["ls"]
        with pytest.raises(ProcessExecutionError) as execinfo:
            ls("-a", "")  # check that empty strings are rendered correctly
        assert execinfo.value.argv[-2:] == ["-a", ""]

    def test_exception_pickling(self):
        import pickle

        with pytest.raises(ProcessExecutionError) as exc_info:
            local.cmd.ls("no-file")
        assert pickle.loads(pickle.dumps(exc_info.value)).argv == exc_info.value.argv

    def test_tempdir(self):
        with local.tempdir() as dir:
            assert dir.is_dir()
            data = b"hello world"
            with open(str(dir / "test.txt"), "wb") as f:
                f.write(data)
            with open(str(dir / "test.txt"), "rb") as f:
                assert f.read() == data

        assert not dir.exists()

    def test_direct_open_tmpdir(self):
        with local.tempdir() as dir:
            assert dir.is_dir()
            data = b"hello world"
            with open(dir / "test.txt", "wb") as f:
                f.write(data)
            with open(dir / "test.txt", "rb") as f:
                assert f.read() == data

        assert not dir.exists()

    def test_read_write_str(self):
        with local.tempdir() as tmp:
            data = "hello world"
            (tmp / "foo.txt").write(data)
            assert (tmp / "foo.txt").read() == data

    def test_read_write_unicode(self):
        with local.tempdir() as tmp:
            data = "hello world"
            (tmp / "foo.txt").write(data)
            assert (tmp / "foo.txt").read() == data

    def test_read_write_bin(self):
        with local.tempdir() as tmp:
            data = b"hello world"
            (tmp / "foo.txt").write(data)
            assert (tmp / "foo.txt").read(mode="rb") == data

    def test_links(self):
        with local.tempdir() as tmp:
            src = tmp / "foo.txt"
            dst1 = tmp / "bar.txt"
            dst2 = tmp / "spam.txt"
            data = "hello world"
            src.write(data)
            src.link(dst1)
            assert data == dst1.read()
            src.symlink(dst2)
            assert data == dst2.read()

    def test_list_processes(self):
        assert list(local.list_processes())

    def test_pgrep(self):
        assert list(local.pgrep("[pP]ython"))

    def _generate_sigint(self):
        with pytest.raises(KeyboardInterrupt):  # noqa: PT012
            if sys.platform == "win32":
                from win32api import GenerateConsoleCtrlEvent

                GenerateConsoleCtrlEvent(0, 0)  # send Ctrl+C to current TTY
            else:
                os.kill(0, signal.SIGINT)
            time.sleep(1)

    @skip_without_tty
    @skip_on_windows
    def test_same_sesion(self):
        from plumbum.cmd import sleep

        p = sleep.popen([1000])
        assert p.poll() is None
        self._generate_sigint()
        time.sleep(1)
        assert p.poll() is not None

    @skip_without_tty
    def test_new_session(self):
        from plumbum.cmd import sleep

        p = sleep.popen([1000], new_session=True)
        assert p.poll() is None
        self._generate_sigint()
        time.sleep(1)
        assert p.poll() is None
        p.terminate()

    # Hangs sometimes on Windows
    @skip_on_windows
    @pytest.mark.timeout(20)
    def test_local_daemon(self):
        from plumbum.cmd import sleep

        proc = local.daemonic_popen(sleep[5])
        with pytest.raises(OSError):
            os.waitpid(proc.pid, 0)
        proc.wait()

    @skip_on_windows
    def test_atomic_file(self):
        af1 = AtomicFile("tmp.txt")
        af2 = AtomicFile("tmp.txt")
        af1.write_atomic(b"foo")
        af2.write_atomic(b"bar")
        assert af1.read_atomic() == b"bar"
        assert af2.read_atomic() == b"bar"
        local.path("tmp.txt").delete()

    @skip_on_windows
    def test_atomic_file2(self):
        af = AtomicFile("tmp.txt")

        code = """\
from plumbum.fs.atomic import AtomicFile
af = AtomicFile("tmp.txt")
try:
    with af.locked(blocking = False):
        raise ValueError("this should have failed")
except (OSError, IOError):
    print("already locked")
"""
        with af.locked():
            output = local.python("-c", code)
            assert output.strip() == "already locked"

        local.path("tmp.txt").delete()

    @skip_on_windows
    def test_pid_file(self):
        code = """\
from plumbum.fs.atomic import PidFile, PidFileTaken
try:
    with PidFile("mypid"):
        raise ValueError("this should have failed")
except PidFileTaken:
    print("already locked")
"""
        with PidFile("mypid"):
            output = local.python("-c", code)
            assert output.strip() == "already locked"

        local.path("mypid").delete()

    @skip_on_windows
    def test_atomic_counter(self):
        local.path("counter").delete()
        num_of_procs = 20
        num_of_increments = 20

        code = f"""from plumbum.fs.atomic import AtomicCounterFile
import time
time.sleep(0.2)
afc = AtomicCounterFile.open("counter")
for _ in range({num_of_increments}):
    print(afc.next())
    time.sleep(0.1)
"""

        procs = []
        for _ in range(num_of_procs):
            procs.append(local.python["-c", code].popen())
        results = []
        for p in procs:
            out, _ = p.communicate()
            assert p.returncode == 0
            results.extend(int(num) for num in out.splitlines())

        assert len(results) == num_of_procs * num_of_increments
        assert len(set(results)) == len(results)
        assert min(results) == 0
        assert max(results) == num_of_procs * num_of_increments - 1
        local.path("counter").delete()

    @skip_on_windows
    def test_atomic_counter2(self):
        local.path("counter").delete()
        afc = AtomicCounterFile.open("counter")
        assert afc.next() == 0
        assert afc.next() == 1
        assert afc.next() == 2

        with pytest.raises(TypeError):
            afc.reset("hello")

        afc.reset(70)
        assert afc.next() == 70
        assert afc.next() == 71
        assert afc.next() == 72

        local.path("counter").delete()

    @skip_on_windows
    @pytest.mark.skipif("printenv" not in local, reason="printenv is missing")
    def test_bound_env(self):
        from plumbum.cmd import printenv

        with local.env(FOO="hello"):
            assert printenv.with_env(BAR="world")("FOO") == "hello\n"
            assert printenv.with_env(BAR="world")("BAR") == "world\n"
            assert printenv.with_env(FOO="sea", BAR="world")("FOO") == "sea\n"
            assert printenv("FOO") == "hello\n"

        assert local.cmd.pwd.with_cwd("/")() == "/\n"
        assert local.cmd.pwd["-L"].with_env(A="X").with_cwd("/")() == "/\n"

    def test_nesting_lists_as_argv(self):
        from plumbum.cmd import ls

        c = ls["-l", ["-a", "*.py"]]
        assert c.formulate()[1:] == ["-l", "-a", "*.py"]

    def test_contains_ls(self):
        assert "ls" in local

    def test_issue_139(self):
        LocalPath(local.cwd)

    def test_pipeline_failure(self):
        from plumbum.cmd import head, ls

        with pytest.raises(ProcessExecutionError):
            (ls["--no-such-option"] | head)()

    def test_cmd(self):
        local.cmd.ls("/tmp")

    def test_pipeline_retcode(self):
        "From PR #288"
        from plumbum.cmd import echo, grep

        print((echo["one two three four"] | grep["two"] | grep["three"])(retcode=None))
        print((echo["one two three four"] | grep["five"] | grep["three"])(retcode=None))
        print((echo["one two three four"] | grep["two"] | grep["five"])(retcode=None))
        print((echo["one two three four"] | grep["six"] | grep["five"])(retcode=None))

    def test_pipeline_stdin(self):
        from subprocess import PIPE

        from plumbum.cmd import cat

        with (cat | cat).bgrun(stdin=PIPE) as future:
            future.stdin.write(b"foobar")
            future.stdin.close()

    def test_run_bg(self):
        from plumbum.cmd import ls

        f = ls["-a"].run_bg()
        f.wait()
        assert "test_local.py" in f.stdout

    def test_run_fg(self, capfd):
        from plumbum.cmd import ls

        ls["-l"].run_fg()
        stdout = capfd.readouterr()[0]
        assert "test_local.py" in stdout

    @skip_on_windows
    def test_run_tee(self, capfd):
        from plumbum.cmd import echo

        result = echo["This is fun"].run_tee()
        assert result[1] == "This is fun\n"
        assert capfd.readouterr()[0] == "This is fun\n"

    def test_run_tf(self):
        from plumbum.cmd import ls

        f = ls["-l"].run_tf()
        assert f is True

    def test_run_retcode(self):
        from plumbum.cmd import ls

        f = ls["-l"].run_retcode()
        assert f == 0

    def test_run_nohup(self):
        from plumbum.cmd import ls

        f = ls["-l"].run_nohup()
        f.wait()
        assert os.path.exists("nohup.out")
        os.unlink("nohup.out")


class TestLocalEncoding:
    try:
        richstr = unichr(40960)
    except NameError:
        richstr = chr(40960)

    def test_inout_rich(self):
        from plumbum.cmd import echo

        out = echo(self.richstr)
        assert self.richstr in out

    @pytest.mark.usefixtures("cleandir")
    def test_out_rich(self):
        from plumbum.cmd import cat

        with open("temp.txt", "w", encoding="utf8") as f:
            f.write(self.richstr)
        out = cat("temp.txt")
        assert self.richstr in out

    @pytest.mark.xfail(IS_WIN32, reason="Unicode path not supported on Windows for now")
    @pytest.mark.usefixtures("cleandir")
    def test_runfile_rich(self):
        import os
        import stat

        name = self.richstr + "_program"
        with open(name, "w") as f:
            f.write(f"#!{sys.executable}\nprint('yes')")

        st = os.stat(name)
        os.chmod(name, st.st_mode | stat.S_IEXEC)

        assert "yes" in local[local.cwd / name]()


@pytest.mark.skipif(
    IS_WIN32, reason="Windows does not support these weird paths, so unambiguous there"
)
def test_local_glob_path(tmpdir):
    p = tmpdir.mkdir("a*b?c")
    p2 = tmpdir.mkdir("aanythingbxc")
    p2.join("something.txt").write("content")
    p.join("hello.txt").write("content")
    p.join("other.txt").write("content")

    pp = LocalPath(str(p))
    assert len(pp // "*.txt") == 2



from __future__ import annotations

from plumbum import cli, local
from plumbum.cli.terminal import get_terminal_size


class SimpleApp(cli.Application):
    @cli.switch(["a"])
    def spam(self):
        print("!!a")

    @cli.switch(
        ["b", "bacon"], argtype=int, mandatory=True, envname="PLUMBUM_TEST_BACON"
    )
    def bacon(self, param):
        """give me some bacon"""
        print("!!b", param)

    eggs = cli.SwitchAttr(
        ["e"], str, help="sets the eggs attribute", envname="PLUMBUM_TEST_EGGS"
    )
    cheese = cli.Flag(["--cheese"], help="cheese, please")
    chives = cli.Flag(["--chives"], help="chives, instead")
    verbose = cli.CountOf(["v"], help="increases the verbosity level")
    benedict = cli.CountOf(
        ["--benedict"],
        help="""a very long help message with lots of
        useless information that nobody would ever want to read, but heck, we need to test
        text wrapping in help messages as well""",
    )

    csv = cli.SwitchAttr(
        ["--csv"], cli.Set("MIN", "MAX", int, csv=True, all_markers={"all"})
    )
    num = cli.SwitchAttr(["--num"], cli.Set("MIN", "MAX", int))

    def main(self, *args):
        old = self.eggs
        self.eggs = "lalala"
        self.eggs = old
        self.tailargs = args

        print(self.csv)


class PositionalApp(cli.Application):
    def main(self, one):
        print("Got", one)


class Geet(cli.Application):
    debug = cli.Flag("--debug")
    cleanups = []

    def main(self):
        del self.cleanups[:]
        print("hi this is geet main")

    def cleanup(self, retcode):
        self.cleanups.append(1)
        print(f"geet cleaning up with rc = {retcode}")


@Geet.subcommand("add")
class GeetAdd(cli.Application):
    def main(self, *files):
        return "adding", files


@Geet.subcommand("commit")
class GeetCommit(cli.Application):
    message = cli.Flag("-m", str)

    def main(self):
        if self.parent.debug:
            return "committing in debug"
        return "committing"

    def cleanup(self, retcode):
        self.parent.cleanups.append(2)
        print(f"geet commit cleaning up with rc = {retcode}")


class Sample(cli.Application):
    DESCRIPTION = "A sample cli application"
    DESCRIPTION_MORE = """
    ABC This is just a sample help text typed with a Dvorak keyboard.
 Although this paragraph is not left or right justified
      in source, we expect it to appear
   formatted nicely on the output, maintaining the indentation of the first line.

  DEF this one has a different indentation.

Let's test that list items are not combined as paragraphs.

   - Item 1
  GHI more text for item 1, which may be very very very very very very long and even more long and long and long to
     prove that we can actually wrap list items as well.
   - Item 2 and this is
     some text for item 2
   - Item 3

List items with invisible bullets should be printed without the bullet.

 /XYZ Invisible 1
 /Invisible 2

  * Star 1
  * Star 2

  Last paragraph can fill more than one line on the output as well. So many features is bound to cause lots of bugs.
  Oh well...
    """

    foo = cli.SwitchAttr("--foo")


Sample.unbind_switches("--version")


class Mumble(cli.Application):
    pass


Sample.subcommand("mumble", Mumble)


class LazyLoaded(cli.Application):
    def main(self):
        print("hello world")


class AppA(cli.Application):
    @cli.switch(["--one"])
    def one(self):
        pass

    two = cli.SwitchAttr(["--two"])


class AppB(AppA):
    @cli.switch(["--three"])
    def three(self):
        pass

    four = cli.SwitchAttr(["--four"])

    def main(self):
        pass


# Testing #363
class TestInheritedApp:
    def test_help(self, capsys):
        _, rc = AppB.run(["AppB", "-h"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()
        assert "--one" in stdout
        assert "--two" in stdout
        assert "--three" in stdout
        assert "--four" in stdout


class TestCLI:
    def test_meta_switches(self):
        _, rc = SimpleApp.run(["foo", "-h"], exit=False)
        assert rc == 0
        _, rc = SimpleApp.run(["foo", "--version"], exit=False)
        assert rc == 0

    def test_okay(self, capsys):
        _, rc = SimpleApp.run(["foo", "--bacon=81"], exit=False)
        assert rc == 0

        inst, rc = SimpleApp.run(
            [
                "foo",
                "--bacon=81",
                "-a",
                "-v",
                "-e",
                "7",
                "-vv",
                "--",
                "lala",
                "-e",
                "7",
            ],
            exit=False,
        )
        assert rc == 0
        assert inst.eggs == "7"

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--csv=100"], exit=False)
        assert rc == 0

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--csv=MAX,MIN,100"], exit=False)
        assert rc == 0

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--num=100"], exit=False)
        assert rc == 0

        capsys.readouterr()
        _, rc = SimpleApp.run(["foo", "--bacon=81", "--csv=all,100"], exit=False)
        assert rc == 0
        output = capsys.readouterr()
        assert "min" in output.out
        assert "max" in output.out
        assert "100" in output.out

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--num=MAX"], exit=False)
        assert rc == 0

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--num=MIN"], exit=False)
        assert rc == 0

    def test_failures(self):
        _, rc = SimpleApp.run(["foo"], exit=False)
        assert rc == 2

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--csv=xx"], exit=False)
        assert rc == 2

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--csv=xx"], exit=False)
        assert rc == 2

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--num=MOO"], exit=False)
        assert rc == 2

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--num=MIN,MAX"], exit=False)
        assert rc == 2

        _, rc = SimpleApp.run(["foo", "--bacon=81", "--num=10.5"], exit=False)
        assert rc == 2

        _, rc = SimpleApp.run(["foo", "--bacon=hello"], exit=False)
        assert rc == 2

    # Testing #371
    def test_extra_args(self, capsys):
        _, rc = PositionalApp.run(["positionalapp"], exit=False)
        assert rc != 0
        stdout, _ = capsys.readouterr()
        assert "Expected at least" in stdout

        _, rc = PositionalApp.run(["positionalapp", "one"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()

        _, rc = PositionalApp.run(["positionalapp", "one", "two"], exit=False)
        assert rc != 0
        stdout, _ = capsys.readouterr()
        assert "Expected at most" in stdout

    def test_subcommands(self):
        _, rc = Geet.run(["geet", "--debug"], exit=False)
        assert rc == 0
        assert Geet.cleanups == [1]
        _, rc = Geet.run(["geet", "--debug", "add", "foo.txt", "bar.txt"], exit=False)
        assert rc == ("adding", ("foo.txt", "bar.txt"))
        assert Geet.cleanups == [1]
        _, rc = Geet.run(["geet", "--debug", "commit"], exit=False)
        assert rc == "committing in debug"
        assert Geet.cleanups == [2, 1]
        _, rc = Geet.run(["geet", "--help"], exit=False)
        assert rc == 0
        _, rc = Geet.run(["geet", "commit", "--help"], exit=False)
        assert rc == 0
        assert Geet.cleanups == [1]

    def test_help_all(self, capsys):
        _, rc = Geet.run(["geet", "--help-all"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()
        assert "--help-all" in stdout
        assert "geet add" in stdout
        assert "geet commit" in stdout

    def test_unbind(self, capsys):
        _, rc = Sample.run(["sample", "--help"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()
        assert "--foo" in stdout
        assert "--version" not in stdout

    def test_description(self, capsys):
        _, rc = Sample.run(["sample", "--help"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()
        cols, _ = get_terminal_size()

        if cols < 9:
            # Terminal is too narrow to test
            pass
        else:
            # Paragraph indentation should be preserved
            assert "    ABC" in stdout
            assert "  DEF" in stdout
            assert "   - Item" in stdout
            # List items should not be combined into paragraphs
            assert "  * Star 2" in stdout
            # Lines of the same list item should be combined. (The right-hand expression of the 'or' operator
            # below is for when the terminal is too narrow, causing "GHI" to be wrapped to the next line.)
            assert "  GHI" not in stdout or "     GHI" in stdout
            # List item with invisible bullet should be indented without the bullet
            assert " XYZ" in stdout

    def test_default_main(self, capsys):
        _, rc = Sample.run(["sample"], exit=False)
        assert rc == 1
        stdout, _ = capsys.readouterr()
        assert "No sub-command given" in stdout

        _, rc = Sample.run(["sample", "pimple"], exit=False)
        assert rc == 1
        stdout, _ = capsys.readouterr()
        assert "Unknown sub-command 'pimple'" in stdout

        _, rc = Sample.run(["sample", "mumble"], exit=False)
        assert rc == 1
        stdout, _ = capsys.readouterr()
        assert "main() not implemented" in stdout

    def test_lazy_subcommand(self, capsys):
        class Foo(cli.Application):
            pass

        Foo.subcommand("lazy", "test_cli.LazyLoaded")

        _, rc = Foo.run(["foo", "lazy"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()
        assert "hello world" in stdout

    def test_multiple_subcommand_names(self):
        """Test that a single subapp can be assigned to multiple subcommand names"""

        class MainApp(cli.Application):
            pass

        @MainApp.subcommand("new-name")
        @MainApp.subcommand("legacy-name")
        class SubApp(cli.Application):
            def main(self):
                return "SubApp executed"

        # Test that both names are registered
        _, rc = MainApp.run(["mainapp", "new-name"], exit=False)
        assert rc == "SubApp executed"

        _, rc = MainApp.run(["mainapp", "legacy-name"], exit=False)
        assert rc == "SubApp executed"

        # Test using loop registration (v2 style)
        class AnotherApp(cli.Application):
            pass

        class AnotherSub(cli.Application):
            def main(self):
                return "AnotherSub executed"

        for name in ("alias1", "alias2", "alias3"):
            AnotherApp.subcommand(name, AnotherSub)

        _, rc = AnotherApp.run(["anotherapp", "alias1"], exit=False)
        assert rc == "AnotherSub executed"

        _, rc = AnotherApp.run(["anotherapp", "alias2"], exit=False)
        assert rc == "AnotherSub executed"

        _, rc = AnotherApp.run(["anotherapp", "alias3"], exit=False)
        assert rc == "AnotherSub executed"

    def test_reset_switchattr(self):
        inst, rc = SimpleApp.run(["foo", "--bacon=81", "-e", "bar"], exit=False)
        assert rc == 0
        assert inst.eggs == "bar"

        inst, rc = SimpleApp.run(["foo", "--bacon=81"], exit=False)
        assert rc == 0
        assert inst.eggs is None

    def test_invoke(self):
        inst, _ = SimpleApp.invoke("arg1", "arg2", eggs="sunny", bacon=10, verbose=2)
        assert (inst.eggs, inst.verbose, inst.tailargs) == (
            "sunny",
            2,
            ("arg1", "arg2"),
        )

    def test_invoke_flag(self):
        # Test that Flag kwargs work correctly with invoke()
        # When debug=False, flag should remain False (not toggled)
        inst, _ = Geet.invoke(debug=False)
        assert inst.debug is False

        # When debug=True, flag should be True
        inst, _ = Geet.invoke(debug=True)
        assert inst.debug is True

        # When no argument, flag should use default (False)
        inst, _ = Geet.invoke()
        assert inst.debug is False

    def test_env_var(self, capsys):
        _, rc = SimpleApp.run(["arg", "--bacon=10"], exit=False)
        assert rc == 0
        stdout, _ = capsys.readouterr()
        assert "10" in stdout

        with local.env(
            PLUMBUM_TEST_BACON="20",
            PLUMBUM_TEST_EGGS="raw",
        ):
            inst, rc = SimpleApp.run(["arg"], exit=False)

        assert rc == 0
        stdout, _ = capsys.readouterr()
        assert "20" in stdout
        assert inst.eggs == "raw"

    def test_mandatory_env_var(self, capsys):
        _, rc = SimpleApp.run(["arg"], exit=False)
        assert rc == 2
        stdout, _ = capsys.readouterr()
        assert "bacon is mandatory" in stdout

    def test_partial_switches(self, capsys):
        app = SimpleApp
        app.ALLOW_ABBREV = True
        inst, rc = app.run(["foo", "--bacon=2", "--ch"], exit=False)
        stdout, _stderr = capsys.readouterr()
        assert "Ambiguous partial switch" in stdout
        assert rc == 2

        inst, rc = app.run(["foo", "--bacon=2", "--chee"], exit=False)
        assert rc == 0
        assert inst.cheese is True
        assert inst.chives is False



from __future__ import annotations

import contextlib

from plumbum import local
from plumbum._testtools import skip_on_windows

with contextlib.suppress(ModuleNotFoundError):
    from plumbum.cmd import printenv


@skip_on_windows
class TestEnv:
    def test_change_env(self):
        with local.env(silly=12):
            assert local.env["silly"] == 12
            actual = {x.split("=")[0] for x in printenv().splitlines() if "=" in x}
            localenv = {x[0] for x in local.env}
            print(actual, localenv)
            assert localenv == actual
            assert len(local.env) == len(actual)

    def test_dictlike(self):
        keys = {x.split("=")[0] for x in printenv().splitlines() if "=" in x}
        values = {
            x.split("=", 1)[1].strip() for x in printenv().splitlines() if "=" in x
        }

        assert keys == set(local.env.keys())
        assert len(values) == len(set(local.env.values()))

    def test_custom_env(self):
        with local.env():
            items = {"one": "OnE", "tww": "TWOO"}
            local.env.update(items)
            assert "tww" in local.env
            local.env.clear()
            assert "tww" not in local.env

    def test_item(self):
        with local.env():
            local.env["simple_plum"] = "thing"
            assert "simple_plum" in local.env
            del local.env["simple_plum"]
            assert "simple_plum" not in local.env
            local.env["simple_plum"] = "thing"
            assert "simple_plum" in local.env
            assert local.env.pop("simple_plum") == "thing"
            assert "simple_plum" not in local.env
            local.env["simple_plum"] = "thing"
        assert "simple_plum" not in local.env

    @skip_on_windows
    def test_home(self):
        assert local.env.home == local.env["HOME"]
        old_home = local.env.home
        with local.env():
            local.env.home = "Nobody"
            assert local.env.home == local.env["HOME"]
            assert local.env.home == "Nobody"
        assert local.env.home == old_home

    @skip_on_windows
    def test_user(self):
        assert local.env.user



from __future__ import annotations

from plumbum import local
from plumbum.paramiko_machine import ParamikoMachine as PM

local.env.path.append("c:\\progra~1\\git\\bin")
from plumbum.cmd import grep, ls  # noqa: E402

m = PM("192.168.1.143")
mls = m["ls"]
mgrep = m["grep"]
# (mls | mgrep["b"])()

(mls | grep["\\."])()

(ls | mgrep["\\."])()



"""
This is provided as a quick way to recover your terminal. Simply run
``python3 -m plumbum.colorlib``
to recover terminal color.
"""

from __future__ import annotations

from . import main

main()



# Just check to see if this file is importable
from __future__ import annotations

from plumbum.cli.image import Image  # noqa: F401
from plumbum.colorlib.names import FindNearest, color_html
from plumbum.colorlib.styles import (  # noqa: F401
    ANSIStyle,
    AttributeNotFound,
    Color,
    ColorNotFound,
)


class TestNearestColor:
    def test_exact(self):
        assert FindNearest(0, 0, 0).all_fast() == 0
        for n, color in enumerate(color_html):
            # Ignoring duplicates
            if n not in (16, 21, 46, 51, 196, 201, 226, 231, 244):
                rgb = (int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16))
                assert FindNearest(*rgb).all_fast() == n

    def test_nearby(self):
        assert FindNearest(1, 2, 2).all_fast() == 0
        assert FindNearest(7, 7, 9).all_fast() == 232

    def test_simplecolor(self):
        assert FindNearest(1, 2, 4).only_basic() == 0
        assert FindNearest(0, 255, 0).only_basic() == 2
        assert FindNearest(100, 100, 0).only_basic() == 3
        assert FindNearest(140, 140, 140).only_basic() == 7


class TestColorLoad:
    def test_rgb(self):
        blue = Color(0, 0, 255)  # Red, Green, Blue
        assert blue.rgb == (0, 0, 255)

    def test_simple_name(self):
        green = Color.from_simple("green")
        assert green.number == 2

    def test_different_names(self):
        assert Color("Dark Blue") == Color("Dark_Blue")
        assert Color("Dark_blue") == Color("Dark_Blue")
        assert Color("DARKBLUE") == Color("Dark_Blue")
        assert Color("DarkBlue") == Color("Dark_Blue")
        assert Color("Dark Green") == Color("Dark_Green")

    def test_loading_methods(self):
        assert Color("Yellow") == Color.from_full("Yellow")
        assert (
            Color.from_full("yellow").representation
            != Color.from_simple("yellow").representation
        )


class TestANSIColor:
    @classmethod
    def setup_class(cls):
        ANSIStyle.use_color = True

    def test_ansi(self):
        assert str(ANSIStyle(fgcolor=Color("reset"))) == "\033[39m"
        assert str(ANSIStyle(fgcolor=Color.from_full("green"))) == "\033[38;5;2m"
        assert str(ANSIStyle(fgcolor=Color.from_simple("red"))) == "\033[31m"


class TestNearestColorAgain:
    def test_allcolors(self):
        myrange = (
            0,
            1,
            2,
            5,
            17,
            39,
            48,
            73,
            82,
            140,
            193,
            210,
            240,
            244,
            250,
            254,
            255,
        )
        for r in myrange:
            for g in myrange:
                for b in myrange:
                    near = FindNearest(r, g, b)
                    assert near.all_slow() == near.all_fast(), f"Tested: {r}, {g}, {b}"



from __future__ import annotations

import functools
import inspect
import os
import sys
import typing
from collections import defaultdict
from textwrap import TextWrapper

from plumbum import colors, local
from plumbum.cli.i18n import get_translation_for
from plumbum.lib import getdoc

from .switches import (
    CountOf,
    Flag,
    MissingArgument,
    MissingMandatorySwitch,
    PositionalArgumentsError,
    SubcommandError,
    SwitchCombinationError,
    SwitchError,
    UnknownSwitch,
    WrongArgumentType,
    switch,
)
from .terminal import get_terminal_size

_translation = get_translation_for(__name__)
T_, ngettext = _translation.gettext, _translation.ngettext


class ShowHelp(SwitchError):
    pass


class ShowHelpAll(SwitchError):
    pass


class ShowVersion(SwitchError):
    pass


class SwitchParseInfo:
    __slots__ = ["__weakref__", "index", "swname", "val"]

    def __init__(self, swname, val, index):
        self.swname = swname
        self.val = val
        self.index = index


class Subcommand:
    def __init__(self, name, subapplication):
        self.name = name
        self.subapplication = subapplication

    def get(self):
        if isinstance(self.subapplication, str):
            modname, clsname = self.subapplication.rsplit(".", 1)
            mod = __import__(modname, None, None, "*")
            try:
                cls = getattr(mod, clsname)
            except AttributeError:
                raise ImportError(f"cannot import name {clsname}") from None
            self.subapplication = cls
        return self.subapplication

    def __repr__(self):
        return T_("Subcommand({self.name}, {self.subapplication})").format(self=self)


_switch_groups = ["Switches", "Meta-switches"]
_switch_groups_l10n = [T_("Switches"), T_("Meta-switches")]


# ===================================================================================================
# CLI Application base class
# ===================================================================================================


class Application:
    """The base class for CLI applications; your "entry point" class should derive from it,
    define the relevant switch functions and attributes, and the ``main()`` function.
    The class defines two overridable "meta switches" for version (``-v``, ``--version``)
    help (``-h``, ``--help``), and help-all (``--help-all``).

    The signature of the main function matters: any positional arguments (e.g., non-switch
    arguments) given on the command line are passed to the ``main()`` function; if you wish
    to allow unlimited number of positional arguments, use varargs (``*args``). The names
    of the arguments will be shown in the help message.

    The classmethod ``run`` serves as the entry point of the class. It parses the command-line
    arguments, invokes switch functions and enters ``main``. You should **not override** this
    method.

    Usage::

        class FileCopier(Application):
            stat = Flag("p", "copy stat info as well")

            def main(self, src, dst):
                if self.stat:
                    shutil.copy2(src, dst)
                else:
                    shutil.copy(src, dst)

        if __name__ == "__main__":
            FileCopier.run()

    There are several class-level attributes you may set:

    * ``PROGNAME`` - the name of the program; if ``None`` (the default), it is set to the
      name of the executable (``argv[0]``); can be in color. If only a color, will be applied to the name.

    * ``VERSION`` - the program's version (defaults to ``1.0``, can be in color)

    * ``DESCRIPTION`` - a short description of your program (shown in help). If not set,
      the class' ``__doc__`` will be used. Can be in color.

    * ``DESCRIPTION_MORE`` - a detailed description of your program (shown in help). The text will be printed
      by paragraphs (specified by empty lines between them). The indentation of each paragraph will be the
      indentation of its first line. List items are identified by their first non-whitespace character being
      one of '-', '*', and '/'; so that they are not combined with preceding paragraphs. Bullet '/' is
      "invisible", meaning that the bullet itself will not be printed to the output.

    * ``USAGE`` - the usage line (shown in help).

    * ``COLOR_USAGE_TITLE`` - The color of the usage line's header.

    * ``COLOR_USAGE`` - The color of the usage line.

    * ``COLOR_GROUPS`` - A dictionary that sets colors for the groups, like Meta-switches, Switches,
      and Subcommands.

    * ``COLOR_GROUP_TITLES`` - A dictionary that sets colors for the group titles. If the dictionary is empty,
      it defaults to ``COLOR_GROUPS``.

    * ``SUBCOMMAND_HELPMSG`` - Controls the printing of extra "see subcommand -h" help message.
      Default is a message, set to False to remove.

    * ``ALLOW_ABBREV`` - Controls whether partial switch names are supported, for example '--ver' will match
      '--verbose'. Default is False for backward consistency with previous plumbum releases. Note that ambiguous
      abbreviations will not match, for example if --foothis and --foothat are defined, then --foo will not match.

    A note on sub-commands: when an application is the root, its ``parent`` attribute is set to
    ``None``. When it is used as a nested-command, ``parent`` will point to its direct ancestor.
    Likewise, when an application is invoked with a sub-command, its ``nested_command`` attribute
    will hold the chosen sub-application and its command-line arguments (a tuple); otherwise, it
    will be set to ``None``

    """

    PROGNAME = None
    DESCRIPTION = None
    DESCRIPTION_MORE = None
    VERSION = None
    USAGE = None
    COLOR_USAGE = None
    COLOR_USAGE_TITLE = None
    COLOR_GROUPS = None
    COLOR_GROUP_TITLES = None
    CALL_MAIN_IF_NESTED_COMMAND = True
    SUBCOMMAND_HELPMSG = T_("see '{parent} {sub} --help' for more info")
    ALLOW_ABBREV = False

    parent = None
    nested_command = None
    _unbound_switches = ()

    def __new__(cls, executable=None):
        """Allows running the class directly as a shortcut for main.
        This is necessary for some setup scripts that want a single function,
        instead of an expression with a dot in it."""

        if executable is None:
            # This return value was not a class instance, so __init__ is never called
            return cls.run()

        return super().__new__(cls)

    def __init__(self, executable):
        # Filter colors

        if self.PROGNAME is None:
            self.PROGNAME = os.path.basename(executable)
        elif isinstance(self.PROGNAME, colors._style):
            self.PROGNAME = self.PROGNAME | os.path.basename(executable)
        elif colors.filter(self.PROGNAME) == "":
            self.PROGNAME = colors.extract(self.PROGNAME) | os.path.basename(executable)
        if self.DESCRIPTION is None:
            self.DESCRIPTION = getdoc(self)

        # Allow None for the colors
        self.COLOR_GROUPS = defaultdict(
            lambda: colors.do_nothing,
            {} if type(self).COLOR_GROUPS is None else type(self).COLOR_GROUPS,
        )

        self.COLOR_GROUP_TITLES = defaultdict(
            lambda: colors.do_nothing,
            self.COLOR_GROUPS
            if type(self).COLOR_GROUP_TITLES is None
            else type(self).COLOR_GROUP_TITLES,
        )
        if type(self).COLOR_USAGE is None:
            self.COLOR_USAGE = colors.do_nothing

        self.executable = executable
        self._switches_by_name = {}
        self._switches_by_func = {}
        self._switches_by_envar = {}
        self._subcommands = {}

        for cls in reversed(type(self).mro()):
            for obj in cls.__dict__.values():
                if isinstance(obj, Subcommand):
                    name = colors.filter(obj.name)
                    if name.startswith("-"):
                        raise SubcommandError(
                            T_("Sub-command names cannot start with '-'")
                        )
                    # it's okay for child classes to override sub-commands set by their parents
                    self._subcommands[name] = obj
                    continue

                swinfo = getattr(obj, "_switch_info", None)
                if not swinfo:
                    continue
                for name in swinfo.names:
                    if name in self._unbound_switches:
                        continue
                    if (
                        name in self._switches_by_name
                        and not self._switches_by_name[name].overridable
                    ):
                        raise SwitchError(
                            T_(
                                "Switch {name} already defined and is not overridable"
                            ).format(name=name)
                        )
                    self._switches_by_name[name] = swinfo
                    self._switches_by_func[swinfo.func] = swinfo
                    if swinfo.envname:
                        self._switches_by_envar[swinfo.envname] = swinfo

    @property
    def root_app(self):
        return self.parent.root_app if self.parent else self

    @classmethod
    def unbind_switches(cls, *switch_names):
        """Unbinds the given switch names from this application. For example

        ::

            class MyApp(cli.Application):
                pass
            MyApp.unbind_switches("--version")

        """
        cls._unbound_switches += tuple(
            name.lstrip("-") for name in switch_names if name
        )

    @classmethod
    def subcommand(cls, name, subapp=None):
        """Registers the given sub-application as a sub-command of this one. This method can be
        used both as a decorator and as a normal ``classmethod``::

            @MyApp.subcommand("foo")
            class FooApp(cli.Application):
                pass

        Or ::

            MyApp.subcommand("foo", FooApp)

        .. versionadded:: 1.1

        .. versionadded:: 1.3
            The sub-command can also be a string, in which case it is treated as a
            fully-qualified class name and is imported on demand. For example,

            MyApp.subcommand("foo", "fully.qualified.package.FooApp")

        """

        def wrapper(subapp):
            # Use the subcommand name (not subapp name) to ensure uniqueness
            # This allows the same subapp to be registered under multiple names
            attrname = f"_subcommand_{name}"
            setattr(cls, attrname, Subcommand(name, subapp))
            return subapp

        return wrapper(subapp) if subapp else wrapper

    def _get_partial_matches(self, partialname):
        matches = []
        for switch_ in self._switches_by_name:
            if switch_.startswith(partialname):
                matches += [
                    switch_,
                ]
        return matches

    def _parse_args(self, argv):
        tailargs = []
        swfuncs = {}
        index = 0

        while argv:
            index += 1
            a = argv.pop(0)
            val = None
            if a == "--":
                # end of options, treat the rest as tailargs
                tailargs.extend(argv)
                break

            if a in self._subcommands:
                subcmd = self._subcommands[a].get()
                self.nested_command = (
                    subcmd,
                    [self.PROGNAME + " " + self._subcommands[a].name, *argv],
                )
                break

            if a.startswith("--") and len(a) >= 3:
                # [--name], [--name=XXX], [--name, XXX], [--name, ==, XXX],
                # [--name=, XXX], [--name, =XXX]
                eqsign = a.find("=")
                if eqsign >= 0:
                    name = a[2:eqsign]
                    argv.insert(0, a[eqsign:])
                else:
                    name = a[2:]

                if self.ALLOW_ABBREV:
                    partials = self._get_partial_matches(name)
                    if len(partials) == 1:
                        name = partials[0]
                    elif len(partials) > 1:
                        raise UnknownSwitch(
                            T_("Ambiguous partial switch {0}").format("--" + name)
                        )

                swname = "--" + name
                if name not in self._switches_by_name:
                    raise UnknownSwitch(T_("Unknown switch {0}").format(swname))
                swinfo = self._switches_by_name[name]
                if swinfo.argtype:
                    if not argv:
                        raise MissingArgument(
                            T_("Switch {0} requires an argument").format(swname)
                        )
                    a = argv.pop(0)
                    if a and a[0] == "=":
                        if len(a) >= 2:
                            val = a[1:]
                        else:
                            if not argv:
                                raise MissingArgument(
                                    T_("Switch {0} requires an argument").format(swname)
                                )
                            val = argv.pop(0)
                    else:
                        val = a

            elif a.startswith("-") and len(a) >= 2:
                # [-a], [-a, XXX], [-aXXX], [-abc]
                name = a[1]
                swname = "-" + name
                if name not in self._switches_by_name:
                    raise UnknownSwitch(T_("Unknown switch {0}").format(swname))
                swinfo = self._switches_by_name[name]
                if swinfo.argtype:
                    if len(a) >= 3:
                        val = a[2:]
                    else:
                        if not argv:
                            raise MissingArgument(
                                T_("Switch {0} requires an argument").format(swname)
                            )
                        val = argv.pop(0)
                elif len(a) >= 3:
                    argv.insert(0, "-" + a[2:])

            else:
                if a.startswith("-"):
                    raise UnknownSwitch(T_("Unknown switch {0}").format(a))
                tailargs.append(a)
                continue

            # handle argument
            val = self._handle_argument(val, swinfo.argtype, name)

            if swinfo.func in swfuncs:
                if swinfo.list:
                    swfuncs[swinfo.func].val[0].append(val)
                else:
                    if swfuncs[swinfo.func].swname == swname:
                        raise SwitchError(T_("Switch {0} already given").format(swname))
                    raise SwitchError(
                        T_("Switch {0} already given ({1} is equivalent)").format(
                            swfuncs[swinfo.func].swname, swname
                        )
                    )
            else:
                if swinfo.list:
                    swfuncs[swinfo.func] = SwitchParseInfo(swname, ([val],), index)
                elif val is NotImplemented:
                    swfuncs[swinfo.func] = SwitchParseInfo(swname, (), index)
                else:
                    swfuncs[swinfo.func] = SwitchParseInfo(swname, (val,), index)

        # Extracting arguments from environment variables
        envindex = 0
        for env, swinfo in self._switches_by_envar.items():
            envindex -= 1
            envval = local.env.get(env)
            if envval is None:
                continue

            if swinfo.func in swfuncs:
                continue  # skip if overridden by command line arguments

            val = self._handle_argument(envval, swinfo.argtype, env)
            envname = f"${env}"
            if swinfo.list:
                # multiple values over environment variables are not supported,
                # this will require some sort of escaping and separator convention
                swfuncs[swinfo.func] = SwitchParseInfo(envname, ([val],), envindex)
            elif val is NotImplemented:
                swfuncs[swinfo.func] = SwitchParseInfo(envname, (), envindex)
            else:
                swfuncs[swinfo.func] = SwitchParseInfo(envname, (val,), envindex)

        return swfuncs, tailargs

    @classmethod
    def autocomplete(cls, argv):
        """This is supplied to make subclassing and testing argument completion methods easier"""

    @staticmethod
    def _handle_argument(val, argtype, name):
        if argtype:
            try:
                return argtype(val)
            except (TypeError, ValueError) as ex:
                raise WrongArgumentType(
                    T_(
                        "Argument of {name} expected to be {argtype}, not {val!r}:\n    {ex!r}"
                    ).format(name=name, argtype=argtype, val=val, ex=ex)
                ) from None
        else:
            return NotImplemented

    def _validate_args(self, swfuncs, tailargs):
        if self.help.__func__ in swfuncs:
            raise ShowHelp()
        if self.helpall.__func__ in swfuncs:
            raise ShowHelpAll()
        if self.version.__func__ in swfuncs:
            raise ShowVersion()

        requirements = {}
        exclusions = {}
        for swinfo in self._switches_by_func.values():
            if swinfo.mandatory and swinfo.func not in swfuncs:
                raise MissingMandatorySwitch(
                    T_("Switch {0} is mandatory").format(
                        "/".join(
                            ("-" if len(n) == 1 else "--") + n for n in swinfo.names
                        )
                    )
                )
            requirements[swinfo.func] = {
                self._switches_by_name[req] for req in swinfo.requires
            }
            exclusions[swinfo.func] = {
                self._switches_by_name[exc] for exc in swinfo.excludes
            }

        # TODO: compute topological order

        gotten = set(swfuncs.keys())
        for func in gotten:
            missing = {f.func for f in requirements[func]} - gotten
            if missing:
                raise SwitchCombinationError(
                    T_("Given {0}, the following are missing {1}").format(
                        swfuncs[func].swname,
                        [self._switches_by_func[f].names[0] for f in missing],
                    )
                )
            invalid = {f.func for f in exclusions[func]} & gotten
            if invalid:
                raise SwitchCombinationError(
                    T_("Given {0}, the following are invalid {1}").format(
                        swfuncs[func].swname, [swfuncs[f].swname for f in invalid]
                    )
                )

        m = inspect.getfullargspec(self.main)

        max_args = sys.maxsize if m.varargs else len(m.args) - 1
        min_args = len(m.args) - 1 - (len(m.defaults) if m.defaults else 0)
        if len(tailargs) < min_args:
            raise PositionalArgumentsError(
                ngettext(
                    "Expected at least {0} positional argument, got {1}",
                    "Expected at least {0} positional arguments, got {1}",
                    min_args,
                ).format(min_args, tailargs)
            )
        if len(tailargs) > max_args:
            raise PositionalArgumentsError(
                ngettext(
                    "Expected at most {0} positional argument, got {1}",
                    "Expected at most {0} positional arguments, got {1}",
                    max_args,
                ).format(max_args, tailargs)
            )

        # Positional argument validation
        if hasattr(self.main, "positional"):
            tailargs = self._positional_validate(
                tailargs,
                self.main.positional,
                self.main.positional_varargs,
                m.args[1:],
                m.varargs,
            )

        elif hasattr(m, "annotations") and m.annotations:
            annotations = typing.get_type_hints(self.main)
            args_names = list(m.args[1:])
            positional = [None] * len(args_names)
            varargs = None

            # All args are positional, so convert kargs to positional
            for item, annotation in annotations.items():
                if item == m.varargs:
                    varargs = annotation
                elif item != "return":
                    positional[args_names.index(item)] = annotation

            tailargs = self._positional_validate(
                tailargs, positional, varargs, m.args[1:], m.varargs
            )

        ordered = [
            (f, a)
            for _, f, a in sorted((sf.index, f, sf.val) for f, sf in swfuncs.items())
        ]
        return ordered, tailargs

    def _positional_validate(self, args, validator_list, varargs, argnames, varargname):
        """Makes sure args follows the validation given input"""
        out_args = list(args)

        for i in range(min(len(args), len(validator_list))):
            if validator_list[i] is not None:
                out_args[i] = self._handle_argument(
                    args[i], validator_list[i], argnames[i]
                )

        if len(args) > len(validator_list):
            if varargs is not None:
                out_args[len(validator_list) :] = [
                    self._handle_argument(a, varargs, varargname)
                    for a in args[len(validator_list) :]
                ]
            else:
                out_args[len(validator_list) :] = args[len(validator_list) :]

        return out_args

    @classmethod
    def run(
        cls,
        argv=None,
        exit=True,  # pylint: disable=redefined-builtin
    ):
        """
        Runs the application, taking the arguments from ``sys.argv`` by default if
        nothing is passed. If ``exit`` is
        ``True`` (the default), the function will exit with the appropriate return code;
        otherwise it will return a tuple of ``(inst, retcode)``, where ``inst`` is the
        application instance created internally by this function and ``retcode`` is the
        exit code of the application.

        .. note::
           Setting ``exit`` to ``False`` is intended for testing/debugging purposes only -- do
           not override it in other situations.
        """
        if argv is None:
            argv = sys.argv
        cls.autocomplete(argv)
        argv = list(argv)
        inst = cls(argv.pop(0))
        retcode = 0
        try:
            swfuncs, tailargs = inst._parse_args(argv)
            ordered, tailargs = inst._validate_args(swfuncs, tailargs)
        except ShowHelp:
            inst.help()
        except ShowHelpAll:
            inst.helpall()
        except ShowVersion:
            inst.version()
        except SwitchError as ex:
            print(T_("Error: {0}").format(ex))
            print(T_("------"))
            inst.help()
            retcode = 2
        else:
            for f, a in ordered:
                f(inst, *a)

            cleanup = None
            if not inst.nested_command or inst.CALL_MAIN_IF_NESTED_COMMAND:
                retcode = inst.main(*tailargs)
                cleanup = functools.partial(inst.cleanup, retcode)
            if not retcode and inst.nested_command:
                subapp, argv = inst.nested_command
                subapp.parent = inst
                inst, retcode = subapp.run(argv, exit=False)

            if cleanup:
                cleanup()

            if retcode is None:
                retcode = 0

        if exit:
            sys.exit(retcode)
        else:
            return inst, retcode

    @classmethod
    def invoke(cls, *args, **switches):
        """Invoke this application programmatically (as a function), in the same way ``run()``
        would. There are two key differences: the return value of ``main()`` is not converted to
        an integer (returned as-is), and exceptions are not swallowed either.

        :param args: any positional arguments for ``main()``
        :param switches: command-line switches are passed as keyword arguments,
                         e.g., ``foo=5`` for ``--foo=5``
        """

        inst = cls("")

        swfuncs = inst._parse_kwd_args(switches)
        ordered, tailargs = inst._validate_args(swfuncs, args)
        for f, a in ordered:
            f(inst, *a)

        cleanup = None
        if not inst.nested_command or inst.CALL_MAIN_IF_NESTED_COMMAND:
            retcode = inst.main(*tailargs)
            cleanup = functools.partial(inst.cleanup, retcode)
        if not retcode and inst.nested_command:
            subapp, argv = inst.nested_command
            subapp.parent = inst
            inst, retcode = subapp.run(argv, exit=False)

        if cleanup:
            cleanup()

        return inst, retcode

    def _parse_kwd_args(self, switches):
        """Parses keywords (positional arguments), used by invoke."""
        swfuncs = {}
        for index, (swname, val) in enumerate(switches.items(), 1):
            switch_local = getattr(type(self), swname)
            swinfo = self._switches_by_func[switch_local._switch_info.func]
            if isinstance(switch_local, CountOf):
                p = (range(val),)
            elif swinfo.list and not hasattr(val, "__iter__"):
                raise SwitchError(
                    T_("Switch {0} must be a sequence (iterable)").format(swname)
                )
            elif not swinfo.argtype:
                # a flag
                # If val is False or None, skip adding the flag (treat as not present)
                if val in (False, None):
                    continue
                if val not in (True, Flag):
                    raise SwitchError(T_("Switch {0} is a boolean flag").format(swname))
                p = ()
            else:
                p = (val,)
            swfuncs[swinfo.func] = SwitchParseInfo(swname, p, index)
        return swfuncs

    def main(self, *args):
        """Implement me (no need to call super)"""
        if self._subcommands:
            if args:
                print(T_("Unknown sub-command '{0}'").format(args[0]))
                print(T_("------"))
                self.help()
                return 1
            if not self.nested_command:
                print(T_("No sub-command given"))
                print(T_("------"))
                self.help()
                return 1
            return 0

        print(T_("main() not implemented"))
        return 1

    def cleanup(self, retcode):
        """Called after ``main()`` and all sub-applications have executed, to perform any necessary cleanup.

        :param retcode: the return code of ``main()``
        """

    @switch(
        ["--help-all"],
        overridable=True,
        group="Meta-switches",
        help=T_("""Prints help messages of all sub-commands and quits"""),
    )
    def helpall(self):
        """Prints help messages of all sub-commands and quits"""
        self.help()
        print()

        if self._subcommands:
            for name, subcls in sorted(self._subcommands.items()):
                subapp = (subcls.get())(f"{self.PROGNAME} {name}")
                subapp.parent = self
                for si in subapp._switches_by_func.values():
                    if si.group == "Meta-switches":
                        si.group = "Hidden-switches"
                subapp.helpall()

    @switch(
        ["-h", "--help"],
        overridable=True,
        group="Meta-switches",
        help=T_("""Prints this help message and quits"""),
    )
    def help(self):  # @ReservedAssignment
        """Prints this help message and quits"""
        if self._get_prog_version():
            self.version()
            print()
        if self.DESCRIPTION:
            print(self.DESCRIPTION.strip() + "\n")

        def split_indentation(s):
            """Identifies the initial indentation (all spaces) of the string and returns the indentation as well
            as the remainder of the line.
            """
            i = 0
            while i < len(s) and s[i] == " ":
                i += 1
            return s[:i], s[i:]

        def paragraphs(text):
            """Yields each paragraph of text along with its initial and subsequent indentations to be used by
            textwrap.TextWrapper.

            Identifies list items from their first non-space character being one of bullets '-', '*', and '/'.
            However, bullet '/' is invisible and is removed from the list item.

            :param text: The text to separate into paragraphs
            """

            paragraph = None
            initial_indent = ""
            subsequent_indent = ""

            def current():
                """Yields the current result if present."""
                if paragraph:
                    yield paragraph, initial_indent, subsequent_indent

            for part in text.lstrip("\n").split("\n"):
                indent, line = split_indentation(part)

                if len(line) == 0:
                    # Starting a new paragraph
                    yield from current()
                    yield "", "", ""

                    paragraph = None
                    initial_indent = ""
                    subsequent_indent = ""
                else:
                    # Adding to current paragraph
                    def is_list_item(line):
                        """Returns true if the first element of 'line' is a bullet character."""
                        bullets = ["-", "*", "/"]
                        return line[0] in bullets

                    def has_invisible_bullet(line):
                        """Returns true if the first element of 'line' is the invisible bullet ('/')."""
                        return line[0] == "/"

                    if is_list_item(line):
                        # Done with current paragraph
                        yield from current()

                        if has_invisible_bullet(line):
                            line = line[1:]

                        paragraph = line
                        initial_indent = indent

                        # Calculate extra indentation for subsequent lines of this list item
                        i = 1
                        while i < len(line) and line[i] == " ":
                            i += 1
                        subsequent_indent = indent + " " * i
                    else:
                        if not paragraph:
                            # Start a new paragraph
                            paragraph = line
                            initial_indent = indent
                            subsequent_indent = indent
                        else:
                            # Add to current paragraph
                            paragraph = paragraph + " " + line

            yield from current()

        def wrapped_paragraphs(text, width):
            """Yields each line of each paragraph of text after wrapping them on 'width' number of columns.

            :param text: The text to yield wrapped lines of
            :param width: The width of the wrapped output
            """
            if not text:
                return

            width = max(width, 1)

            for paragraph, initial_indent, subsequent_indent in paragraphs(text):
                wrapper = TextWrapper(
                    width,
                    initial_indent=initial_indent,
                    subsequent_indent=subsequent_indent,
                )
                w = wrapper.wrap(paragraph)
                yield from w
                if len(w) == 0:
                    yield ""

        cols, _ = get_terminal_size()
        for line in wrapped_paragraphs(self.DESCRIPTION_MORE, cols):
            print(line)

        m = inspect.getfullargspec(self.main)
        tailargs = m.args[1:]  # skip self
        if m.defaults:
            for i, d in enumerate(reversed(m.defaults)):
                tailargs[-i - 1] = f"[{tailargs[-i - 1]}={d}]"
        if m.varargs:
            tailargs.append(f"{m.varargs}...")
        tailargs = " ".join(tailargs)

        utc = self.COLOR_USAGE_TITLE or self.COLOR_USAGE
        print(utc | T_("Usage:"))

        with self.COLOR_USAGE:
            if not self.USAGE:
                if self._subcommands:
                    self.USAGE = T_(
                        "    {progname} [SWITCHES] [SUBCOMMAND [SWITCHES]] {tailargs}\n"
                    )
                else:
                    self.USAGE = T_("    {progname} [SWITCHES] {tailargs}\n")
            print(
                self.USAGE.format(
                    progname=colors.filter(self.PROGNAME), tailargs=tailargs
                )
            )

        by_groups = {}
        for si in self._switches_by_func.values():
            if si.group not in by_groups:
                by_groups[si.group] = []
            by_groups[si.group].append(si)

        def switchs(by_groups, show_groups):
            for grp, swinfos in sorted(by_groups.items(), key=lambda item: item[0]):
                if show_groups:
                    lgrp = T_(grp) if grp in _switch_groups else grp
                    print(self.COLOR_GROUP_TITLES[grp] | lgrp + ":")

                for si in sorted(swinfos, key=lambda x: x.names):
                    swnames = ", ".join(
                        ("-" if len(n) == 1 else "--") + n
                        for n in si.names
                        if n in self._switches_by_name
                        and self._switches_by_name[n] == si
                    )
                    if si.argtype:
                        if hasattr(si.argtype, "__name__"):
                            typename = si.argtype.__name__
                        else:
                            typename = str(si.argtype)
                        argtype = f" {si.argname.upper()}:{typename}"
                    else:
                        argtype = ""
                    prefix = swnames + argtype
                    yield si, prefix, self.COLOR_GROUPS[grp]

                if show_groups:
                    print()

        sw_width = (
            max(len(prefix) for si, prefix, color in switchs(by_groups, False)) + 4
        )
        description_indent = "    {0}{1}{2}"
        wrapper = TextWrapper(width=max(cols - min(sw_width, 60), 50) - 6)
        indentation = "\n" + " " * (cols - wrapper.width)

        for switch_info, prefix, color in switchs(by_groups, True):
            help_txt = switch_info.help
            if switch_info.list:
                help_txt += T_("; may be given multiple times")
            if switch_info.mandatory:
                help_txt += T_("; required")
            if switch_info.requires:
                help_txt += T_("; requires {0}").format(
                    ", ".join(
                        (("-" if len(switch) == 1 else "--") + switch)
                        for switch in switch_info.requires
                    )
                )
            if switch_info.excludes:
                help_txt += T_("; excludes {0}").format(
                    ", ".join(
                        (("-" if len(switch) == 1 else "--") + switch)
                        for switch in switch_info.excludes
                    )
                )

            msg = indentation.join(
                wrapper.wrap(" ".join(ln.strip() for ln in help_txt.splitlines()))
            )

            if len(prefix) + wrapper.width >= cols:
                padding = indentation
            else:
                padding = " " * max(cols - wrapper.width - len(prefix) - 4, 1)
            print(description_indent.format(color | prefix, padding, color | msg))

        if self._subcommands:
            gc = self.COLOR_GROUP_TITLES["Sub-commands"]
            print(gc | T_("Sub-commands:"))
            for name, subcls in sorted(self._subcommands.items()):
                with gc:
                    subapp = subcls.get()
                    doc = subapp.DESCRIPTION or getdoc(subapp)
                    if self.SUBCOMMAND_HELPMSG:
                        help_str = doc + "; " if doc else ""
                        help_str += self.SUBCOMMAND_HELPMSG.format(
                            parent=self.PROGNAME, sub=name
                        )
                    else:
                        help_str = doc or ""

                    msg = indentation.join(
                        wrapper.wrap(
                            " ".join(ln.strip() for ln in help_str.splitlines())
                        )
                    )

                    if len(name) + wrapper.width >= cols:
                        padding = indentation
                    else:
                        padding = " " * max(cols - wrapper.width - len(name) - 4, 1)
                    if colors.contains_colors(subcls.name):
                        bodycolor = colors.extract(subcls.name)
                    else:
                        bodycolor = gc

                    print(
                        description_indent.format(
                            subcls.name, padding, bodycolor | colors.filter(msg)
                        )
                    )

    def _get_prog_version(self):
        ver = None
        curr = self
        while curr is not None:
            ver = getattr(curr, "VERSION", None)
            if ver is not None:
                return ver
            curr = curr.parent
        return ver

    @switch(
        ["-v", "--version"],
        overridable=True,
        group="Meta-switches",
        help=T_("""Prints the program's version and quits"""),
    )
    def version(self):
        """Prints the program's version and quits"""
        ver = self._get_prog_version()
        ver_name = ver if ver is not None else T_("(version not set)")
        print(f"{self.PROGNAME} {ver_name}")



#!/usr/bin/env python3
from __future__ import annotations

from plumbum import cli


class App(cli.Application):
    # VERSION = "1.2.3"
    # x = cli.SwitchAttr("--lala")
    y = cli.Flag("-f")

    def main(self, x, y):
        pass


@App.subcommand("bar")
class Bar(cli.Application):
    z = cli.Flag("-z")

    def main(self, z, w):
        pass


if __name__ == "__main__":
    App.run()



from __future__ import annotations

import plumbum


def __getattr__(name: str) -> plumbum.machines.LocalCommand:
    """The module-hack that allows us to use ``from plumbum.cmd import some_program``"""
    try:
        return plumbum.local[name]
    except plumbum.CommandNotFound:
        raise AttributeError(name) from None



"""
Terminal size utility
---------------------
"""

from __future__ import annotations

import contextlib
import os
import platform
import warnings
from struct import Struct

from plumbum import local


def get_terminal_size(default: tuple[int, int] = (80, 25)) -> tuple[int, int]:
    """
    Get width and height of console; works on linux, os x, windows and cygwin

    Adapted from https://gist.github.com/jtriley/1108174
    Originally from: http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
    """
    current_os = platform.system()
    if current_os == "Windows":  # pragma: no cover
        size = _get_terminal_size_windows()
        if not size:
            # needed for window's python in cygwin's xterm!
            size = _get_terminal_size_tput()
    elif current_os in ("Linux", "Darwin", "FreeBSD", "SunOS") or current_os.startswith(
        "CYGWIN"
    ):
        size = _get_terminal_size_linux()

    else:  # pragma: no cover
        warnings.warn(
            "Plumbum does not know the type of the current OS for term size, defaulting to UNIX",
            stacklevel=2,
        )
        size = _get_terminal_size_linux()

    # we'll assume the standard 80x25 if for any reason we don't know the terminal size
    if size is None:
        return default
    return size


def _get_terminal_size_windows():  # pragma: no cover
    try:
        from ctypes import create_string_buffer, windll

        STDERR_HANDLE = -12
        h = windll.kernel32.GetStdHandle(STDERR_HANDLE)
        csbi_struct = Struct("hhhhHhhhhhh")
        csbi = create_string_buffer(csbi_struct.size)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            _, _, _, _, _, left, top, right, bottom, _, _ = csbi_struct.unpack(csbi.raw)
            return right - left + 1, bottom - top + 1
        return None
    except Exception:
        return None


def _get_terminal_size_tput():  # pragma: no cover
    # get terminal width
    # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
    try:
        tput = local["tput"]
        cols = int(tput("cols"))
        rows = int(tput("lines"))
        return (cols, rows)
    except Exception:
        return None


def _ioctl_GWINSZ(fd: int) -> tuple[int, int] | None:
    # struct winsize has 4 unsigned shorts: ws_row, ws_col, ws_xpixel, ws_ypixel
    # We only need the first two (rows and cols), but must provide full struct size
    winsize = Struct("HHHH")  # 4 unsigned shorts = 8 bytes
    try:
        import fcntl
        import termios

        # TODO: Clean this up. Problems could be hidden by the broad except.
        result = fcntl.ioctl(fd, termios.TIOCGWINSZ, b"\x00" * winsize.size)
        rows, cols, _, _ = winsize.unpack(result)
        return rows, cols
    except Exception:
        return None


def _get_terminal_size_linux() -> tuple[int, int] | None:
    cr = _ioctl_GWINSZ(0) or _ioctl_GWINSZ(1) or _ioctl_GWINSZ(2)
    if not cr:
        with contextlib.suppress(Exception):
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = _ioctl_GWINSZ(fd)
            os.close(fd)
    if not cr:
        try:
            cr = (int(os.environ["LINES"]), int(os.environ["COLUMNS"]))
        except Exception:
            return None
    return cr[1], cr[0]



from __future__ import annotations

import sys
from io import StringIO

import IPython.display
from IPython.core.magic import Magics, cell_magic, magics_class, needs_local_scope

valid_choices = [x[8:] for x in dir(IPython.display) if x[:8] == "display_"]


@magics_class
class OutputMagics(Magics):  # pragma: no cover
    @needs_local_scope
    @cell_magic
    def to(self, line, cell, local_ns=None):
        choice = line.strip()
        assert choice in valid_choices, "Valid choices for '%%to' are: " + str(
            valid_choices
        )
        display_fn = getattr(IPython.display, "display_" + choice)

        # Captures stdout and renders it in the notebook
        with StringIO() as out:
            old_out = sys.stdout
            try:
                sys.stdout = out
                exec(cell, self.shell.user_ns, local_ns)  # pylint: disable=exec-used
                out.seek(0)
                display_fn(out.getvalue(), raw=True)
            finally:
                sys.stdout = old_out



"""Test that PuttyMachine initializes its SshMachine correctly"""

from __future__ import annotations

import pytest

from plumbum import PuttyMachine, SshMachine


@pytest.fixture(params=["default", "322"])
def ssh_port(request):
    return request.param


class TestPuttyMachine:
    def test_putty_command(self, mocker, ssh_port):
        local = mocker.patch("plumbum.machines.ssh_machine.local")
        init = mocker.spy(SshMachine, "__init__")
        mocker.patch("plumbum.machines.ssh_machine.BaseRemoteMachine")

        host = mocker.MagicMock()
        user = local.env.user
        port = keyfile = None
        ssh_command = local["plink"]
        scp_command = local["pscp"]
        ssh_opts = ["-ssh"]
        if ssh_port == "default":
            putty_port = None
            scp_opts = ()
        else:
            putty_port = int(ssh_port)
            ssh_opts.extend(["-P", ssh_port])
            scp_opts = ["-P", ssh_port]
        encoding = mocker.MagicMock()
        connect_timeout = 20
        new_session = True

        PuttyMachine(
            host,
            port=putty_port,
            connect_timeout=connect_timeout,
            new_session=new_session,
            encoding=encoding,
        )

        init.assert_called_with(
            mocker.ANY,
            host,
            user,
            port,
            keyfile=keyfile,
            ssh_command=ssh_command,
            scp_command=scp_command,
            ssh_opts=ssh_opts,
            scp_opts=scp_opts,
            encoding=encoding,
            connect_timeout=connect_timeout,
            new_session=new_session,
        )

    def test_putty_str(self, mocker):
        local = mocker.patch("plumbum.machines.ssh_machine.local")
        mocker.patch("plumbum.machines.ssh_machine.BaseRemoteMachine")

        host = mocker.MagicMock()
        user = local.env.user

        machine = PuttyMachine(host)
        assert str(machine) == f"putty-ssh://{user}@{host}"



#!/usr/bin/env -S uv run

# /// script
# dependencies = ["nox>=2025.2.9"]
# ///

from __future__ import annotations

import nox

nox.needs_version = ">=2025.2.9"
nox.options.default_venv_backend = "uv|virtualenv"

PYPROJECT = nox.project.load_toml()
ALL_PYTHONS = nox.project.python_versions(PYPROJECT)


@nox.session(reuse_venv=True)
def lint(session):
    """
    Run the linter.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def pylint(session):
    """
    Run pylint.
    """

    session.install("-e.", "paramiko", "ipython", "pylint")
    session.run("pylint", "plumbum", *session.posargs)


@nox.session(python=ALL_PYTHONS, reuse_venv=True)
def tests(session):
    """
    Run the unit and regular tests.
    """
    test_deps = nox.project.dependency_groups(PYPROJECT, "test")
    session.install("-e.", *test_deps)
    session.run("pytest", *session.posargs, env={"PYTHONTRACEMALLOC": "5"})


@nox.session(reuse_venv=True, default=False)
def docs(session):
    """
    Build the docs. Pass "serve" to serve.
    """

    doc_deps = nox.project.dependency_groups(PYPROJECT, "docs")
    session.install("-e.", *doc_deps)
    session.chdir("docs")
    session.run("sphinx-build", "-M", "html", ".", "_build")

    if session.posargs:
        if "serve" in session.posargs:
            session.log("Launching docs at http://localhost:8000/ - use Ctrl-C to quit")
            session.run("python", "-m", "http.server", "8000", "-d", "_build/html")
        else:
            session.log("Unsupported argument to docs")


@nox.session(default=False)
def build(session):
    """
    Build an SDist and wheel.
    """

    session.install("build")
    session.run("python", "-m", "build")


if __name__ == "__main__":
    nox.main()



from __future__ import annotations

import logging
import os
import socket
import time
from multiprocessing import Queue
from threading import Thread

import env
import pytest

import plumbum
from plumbum import (
    NOHUP,
    CommandNotFound,
    ProcessExecutionError,
    ProcessTimedOut,
    RemotePath,
    SshMachine,
    local,
)
from plumbum._testtools import skip_on_windows, skip_without_chown
from plumbum.machines.session import HostPublicKeyUnknown, IncorrectLogin

try:
    import paramiko
except ImportError:
    paramiko = None
else:
    from plumbum.machines.paramiko_machine import ParamikoMachine


pytestmark = pytest.mark.ssh


def strassert(one, two):
    assert str(one) == str(two)


def assert_is_port(port):
    assert 0 < int(port) < 2**16


# TEST_HOST = "192.168.1.143"
TEST_HOST = "127.0.0.1"
if TEST_HOST not in ("::1", "127.0.0.1", "localhost"):
    plumbum.local.env.path.append("c:\\Program Files\\Git\\bin")


@pytest.fixture(scope="session")
def sshpass():
    try:
        return plumbum.local["sshpass"]
    except CommandNotFound:
        pytest.skip("Test requires sshpass")


@skip_on_windows
def test_connection():
    SshMachine(TEST_HOST)


def test_incorrect_login(sshpass):  # noqa: ARG001
    with pytest.raises(IncorrectLogin):
        SshMachine(
            TEST_HOST,
            password="swordfish",
            ssh_opts=[
                "-o",
                "PubkeyAuthentication=no",
                "-o",
                "PreferredAuthentications=password",
            ],
        )


@pytest.mark.xfail(env.LINUX, reason="TODO: no idea why this fails on linux")
def test_hostpubkey_unknown(sshpass):  # noqa: ARG001
    with pytest.raises(HostPublicKeyUnknown):
        SshMachine(
            TEST_HOST,
            password="swordfish",
            ssh_opts=["-o", "UserKnownHostsFile=/dev/null", "-o", "UpdateHostKeys=no"],
        )


@skip_on_windows
class TestRemotePath:
    def _connect(self):
        return SshMachine(TEST_HOST)

    def test_name(self):
        name = RemotePath(self._connect(), "/some/long/path/to/file.txt").name
        assert isinstance(name, str)
        assert str(name) == "file.txt"

    def test_dirname(self):
        name = RemotePath(self._connect(), "/some/long/path/to/file.txt").dirname
        assert isinstance(name, RemotePath)
        assert str(name) == "/some/long/path/to"

    def test_uri(self):
        p1 = RemotePath(self._connect(), "/some/long/path/to/file.txt")
        assert p1.as_uri("ftp")[:6] == "ftp://"
        assert p1.as_uri("ssh")[:6] == "ssh://"
        assert p1.as_uri()[-27:] == "/some/long/path/to/file.txt"

    def test_stem(self):
        p = RemotePath(self._connect(), "/some/long/path/to/file.txt")
        assert p.stem == "file"
        p = RemotePath(self._connect(), "/some/long/path/")
        assert p.stem == "path"

    def test_suffix(self):
        p1 = RemotePath(self._connect(), "/some/long/path/to/file.txt")
        p2 = RemotePath(self._connect(), "file.tar.gz")
        assert p1.suffix == ".txt"
        assert p1.suffixes == [".txt"]
        assert p2.suffix == ".gz"
        assert p2.suffixes == [".tar", ".gz"]
        strassert(
            p1.with_suffix(".tar.gz"),
            RemotePath(self._connect(), "/some/long/path/to/file.tar.gz"),
        )
        strassert(
            p2.with_suffix(".other"), RemotePath(self._connect(), "file.tar.other")
        )
        strassert(
            p2.with_suffix(".other", 2), RemotePath(self._connect(), "file.other")
        )
        strassert(
            p2.with_suffix(".other", 0),
            RemotePath(self._connect(), "file.tar.gz.other"),
        )
        strassert(
            p2.with_suffix(".other", None), RemotePath(self._connect(), "file.other")
        )

    def test_newname(self):
        p1 = RemotePath(self._connect(), "/some/long/path/to/file.txt")
        p2 = RemotePath(self._connect(), "file.tar.gz")
        strassert(
            p1.with_name("something.tar"),
            RemotePath(self._connect(), "/some/long/path/to/something.tar"),
        )
        strassert(
            p2.with_name("something.tar"), RemotePath(self._connect(), "something.tar")
        )

    @skip_without_chown
    def test_chown(self):
        with self._connect() as rem, rem.tempdir() as dir:
            p = dir / "foo.txt"
            p.write(b"hello")
            # because we're connected to localhost, we expect UID and GID to be the same
            assert p.uid == os.getuid()
            assert p.gid == os.getgid()
            p.chown(p.uid.name)
            assert p.uid == os.getuid()

    def test_parent(self):
        p1 = RemotePath(self._connect(), "/some/long/path/to/file.txt")
        p2 = p1.parent
        assert str(p2) == "/some/long/path/to"

    def test_mkdir(self):
        # (identical to test_local.TestLocalPath.test_mkdir)
        with self._connect() as rem:
            with rem.tempdir() as tmp:
                (tmp / "a").mkdir(exist_ok=False, parents=False)
                assert (tmp / "a").exists()
                assert (tmp / "a").is_dir()
                (tmp / "a").mkdir(exist_ok=True, parents=False)
                (tmp / "a").mkdir(exist_ok=True, parents=True)
                with pytest.raises(OSError):
                    (tmp / "a").mkdir(exist_ok=False, parents=False)
                with pytest.raises(OSError):
                    (tmp / "a").mkdir(exist_ok=False, parents=True)
                (tmp / "b" / "bb").mkdir(exist_ok=False, parents=True)
                assert (tmp / "b" / "bb").exists()
                assert (tmp / "b" / "bb").is_dir()
            assert not tmp.exists()

    @pytest.mark.xfail(
        reason="mkdir's mode argument is not yet implemented for remote paths",
        strict=True,
    )
    def test_mkdir_mode(self):
        # (identical to test_local.TestLocalPath.test_mkdir_mode)
        with self._connect() as rem:
            with rem.tempdir() as tmp:
                # just verify that mode argument works the same way it does for
                # Python's own os.mkdir, which takes into account the umask
                # (different from shell mkdir mode argument!); umask on my
                # system is 022 by default, so 033 is ok for testing this
                try:
                    (tmp / "pb_333").mkdir(exist_ok=False, parents=False, mode=0o333)
                    rem.python(
                        "-c",
                        "import os; os.mkdir({}, 0o333)".format(
                            repr(str(tmp / "py_333"))
                        ),
                    )
                    pb_final_mode = oct((tmp / "pb_333").stat().st_mode)
                    py_final_mode = oct((tmp / "py_333").stat().st_mode)
                    assert pb_final_mode == py_final_mode
                finally:
                    # we have to revert this so the tempdir deletion works
                    if (tmp / "pb_333").exists():
                        (tmp / "pb_333").chmod(0o777)
                    if (tmp / "py_333").exists():
                        (tmp / "py_333").chmod(0o777)
            assert not tmp.exists()

    def test_copy(self):
        """
        tests `RemotePath.copy` for the following scenarios:

            * copying a simple file from `file_a` to `copy_of_a` succeeds

            * copying file `file_a` into a directory `a_dir/copy_of_a` succeeds

            * copying a directory `a_dir` over an existing directory path with
              `override=False` fails

            * copying a directory `a_dir` over an existing directory path with
              `override=True` succeeds
        """

        with self._connect() as rem:
            with rem.tempdir() as tmp:
                # setup a file and make sure it exists...
                (tmp / "file_a").touch()
                assert (tmp / "file_a").exists()
                assert (tmp / "file_a").is_file()

                # setup a directory for copying into...
                (tmp / "a_dir").mkdir(exist_ok=False, parents=False)
                assert (tmp / "a_dir").exists()
                assert (tmp / "a_dir").is_dir()

                # setup a 2nd directory for testing `override=False`
                (tmp / "b_dir").mkdir(exist_ok=False, parents=False)
                assert (tmp / "b_dir").exists()
                assert (tmp / "b_dir").is_dir()

                # copying a simple file
                (tmp / "file_a").copy(tmp / "copy_of_a")
                assert (tmp / "copy_of_a").exists()
                assert (tmp / "copy_of_a").is_file()

                # copying into a directory
                (tmp / "file_a").copy(tmp / "a_dir/copy_of_a")
                assert (tmp / "a_dir/copy_of_a").exists()
                assert (tmp / "a_dir/copy_of_a").is_file()

                # copying a directory on top of an existing directory using
                # `override=False` (should fail with TypeError)
                with pytest.raises(TypeError):
                    (tmp / "a_dir").copy(tmp / "b_dir", override=False)

                # copying a directory on top of an existing directory using
                # `override=True` (should copy transparently)
                (tmp / "a_dir").copy(tmp / "b_dir", override=True)
                assert "copy_of_a" in (tmp / "b_dir")

            assert not tmp.exists()


class BaseRemoteMachineTest:
    TUNNEL_PROG_AF_INET = r"""import sys, socket
s = socket.socket()
s.bind(("", 0))
s.listen(1)
sys.stdout.write("{0}\n".format(s.getsockname()[1]))
sys.stdout.flush()
s2, _ = s.accept()
data = s2.recv(100)
s2.send(b"hello " + data)
s2.close()
s.close()
"""

    TUNNEL_PROG_AF_UNIX = r"""import sys, socket, tempfile
s = socket.socket(family=socket.AF_UNIX)
socket_location = tempfile.NamedTemporaryFile()
socket_location.close()
s.bind(socket_location.name)
s.listen(1)
sys.stdout.write("{0}\n".format(s.getsockname()))
sys.stdout.flush()
s2, _ = s.accept()
data = s2.recv(100)
s2.send(b"hello " + data)
s2.close()
s.close()
"""

    def test_basic(self):
        with self._connect() as rem:
            r_ssh = rem["ssh"]
            r_ls = rem["ls"]
            r_grep = rem["grep"]

            lines = r_ls("-a").splitlines()
            assert ".bashrc" in lines or ".bash_profile" in lines
            with rem.cwd(os.path.dirname(os.path.abspath(__file__))):
                cmd = r_ssh[
                    "localhost", "cd", rem.cwd, "&&", r_ls, "|", r_grep["\\.py"]
                ]
                assert "'|'" in str(cmd)
                assert "test_remote.py" in cmd()
                assert "test_remote.py" in [f.name for f in rem.cwd // "*.py"]

    # Testing for #271
    def test_double_chdir(self):
        with self._connect() as rem:
            with rem.cwd(os.path.dirname(os.path.abspath(__file__))):
                rem["ls"]()
            with rem.cwd("/tmp"):
                rem["pwd"]()

    def test_glob(self):
        with (
            self._connect() as rem,
            rem.cwd(os.path.dirname(os.path.abspath(__file__))),
        ):
            filenames = [f.name for f in rem.cwd // ("*.py", "*.bash")]
            assert "test_remote.py" in filenames
            assert "slow_process.bash" in filenames

    def test_glob_spaces(self):
        with (
            self._connect() as rem,
            rem.cwd(os.path.dirname(os.path.abspath(__file__))),
        ):
            filenames = [f.name for f in rem.cwd // ("*space.txt")]
            assert "file with space.txt" in filenames

            filenames = [f.name for f in rem.cwd // ("*with space.txt")]
            assert "file with space.txt" in filenames

    def test_cmd(self):
        with self._connect() as rem:
            rem.cmd.ls("/tmp")

    @pytest.mark.usefixtures("testdir")
    def test_download_upload(self):
        with self._connect() as rem:
            rem.upload("test_remote.py", "/tmp")
            r_ls = rem["ls"]
            r_rm = rem["rm"]
            assert "test_remote.py" in r_ls("/tmp").splitlines()
            rem.download("/tmp/test_remote.py", "/tmp/test_download.txt")
            r_rm("/tmp/test_remote.py")
            r_rm("/tmp/test_download.txt")

    def test_session(self):
        with self._connect() as rem:
            sh = rem.session()
            for _ in range(4):
                _, out, _ = sh.run("ls -a")
                assert ".bashrc" in out or ".bash_profile" in out

    @pytest.mark.xfail(env.PYPY, reason="PyPy sometimes fails here", strict=False)
    def test_env(self):
        with self._connect() as rem:
            with pytest.raises(ProcessExecutionError):
                rem.python("-c", "import os;os.environ['FOOBAR72']")
            with rem.env(FOOBAR72="lala"):
                with rem.env(FOOBAR72="baba"):
                    out = rem.python("-c", "import os;print(os.environ['FOOBAR72'])")
                    assert out.strip() == "baba"
                out = rem.python("-c", "import os;print(os.environ['FOOBAR72'])")
                assert out.strip() == "lala"

            # path manipulation
            with pytest.raises(CommandNotFound):
                rem.which("dummy-executable")
            with rem.cwd(os.path.dirname(os.path.abspath(__file__))):
                rem.env.path.insert(0, rem.cwd / "not-in-path")
                p = rem.which("dummy-executable")
                assert p == rem.cwd / "not-in-path" / "dummy-executable"

    @pytest.mark.xfail(env.PYPY, reason="PyPy sometimes fails here", strict=False)
    @pytest.mark.parametrize(
        "env",
        [
            "lala",
            "-Wl,-O2 -Wl,--sort-common",
            "{{}}",
            "''",
            "!@%_-+=:",
            "'",
            "`",
            "$",
            "\\",
        ],
    )
    def test_env_special_characters(self, env):
        with self._connect() as rem:
            with pytest.raises(ProcessExecutionError):
                rem.python("-c", "import os;print(os.environ['FOOBAR72'])")
            rem.env["FOOBAR72"] = env
            out = rem.python("-c", "import os;print(os.environ['FOOBAR72'])")
            assert out.strip() == env

    def test_read_write(self):
        with self._connect() as rem:
            with rem.tempdir() as dir:
                assert dir.is_dir()
                data = b"hello world"
                (dir / "foo.txt").write(data)
                assert (dir / "foo.txt").read() == data

            assert not dir.exists()

    def test_contains(self):
        with self._connect() as rem:
            assert "ls" in rem

    def test_iter_lines_timeout(self):
        with self._connect() as rem:
            try:
                for i, (out, err) in enumerate(  # noqa: B007
                    rem["ping"]["-i", 0.5, "127.0.0.1"].popen().iter_lines(timeout=4)
                ):
                    print("out:", out)
                    print("err:", err)
            except NotImplementedError as err:
                pytest.skip(str(err))
            except ProcessTimedOut:
                assert i > 3
            else:
                pytest.fail("Expected a timeout")

    def test_iter_lines_error(self):
        with self._connect() as rem:
            with pytest.raises(ProcessExecutionError) as ex:  # noqa: PT012
                for i, _lines in enumerate(rem["ls"]["--bla"].popen()):  # noqa: B007
                    pass
                assert i == 1
            assert "ls: " in ex.value.stderr

    def test_touch(self):
        with self._connect() as rem:
            rfile = rem.cwd / "sillyfile"
            assert not rfile.exists()
            rfile.touch()
            assert rfile.exists()
            rfile.delete()


def serve_reverse_tunnel(queue, port):
    s = socket.socket()
    s.bind(("", port))
    s.listen(1)
    s2, _ = s.accept()
    data = s2.recv(100).decode("ascii").strip()
    queue.put(data)
    s2.close()
    s.close()


@skip_on_windows
class TestRemoteMachine(BaseRemoteMachineTest):
    def _connect(self):
        return SshMachine(TEST_HOST)

    @pytest.mark.parametrize("dynamic_lport", [False, True])
    def test_tunnel(self, dynamic_lport):
        for tunnel_prog in (self.TUNNEL_PROG_AF_INET, self.TUNNEL_PROG_AF_UNIX):
            with self._connect() as rem:
                p = (rem.python["-u"] << tunnel_prog).popen()
                port_or_socket = p.stdout.readline().decode("ascii").strip()
                try:
                    port_or_socket = int(port_or_socket)
                    dhost = "localhost"
                except ValueError:
                    dhost = None

                lport = 12222 if not dynamic_lport else 0

                with rem.tunnel(lport, port_or_socket, dhost=dhost) as tun:
                    if not dynamic_lport:
                        assert tun.lport == lport
                    else:
                        assert_is_port(tun.lport)
                    assert tun.dport == port_or_socket
                    assert not tun.reverse

                    s = socket.socket()
                    s.connect(("localhost", tun.lport))
                    s.send(b"world")
                    data = s.recv(100)
                    s.close()

                print(p.communicate())
                assert data == b"hello world"

    @pytest.mark.parametrize("dynamic_dport", [False, True])
    def test_reverse_tunnel(self, dynamic_dport):
        lport = 12223 + dynamic_dport
        with self._connect() as rem:
            queue = Queue()
            tunnel_server = Thread(target=serve_reverse_tunnel, args=(queue, lport))
            tunnel_server.start()
            message = str(time.time())

            if not dynamic_dport:
                get_unbound_socket_remote = """import sys, socket
s = socket.socket()
s.bind(("", 0))
s.listen(1)
sys.stdout.write(str(s.getsockname()[1]))
sys.stdout.flush()
s.close()
"""
                p = (rem.python["-u"] << get_unbound_socket_remote).popen()
                remote_socket = p.stdout.readline().decode("ascii").strip()
            else:
                remote_socket = 0

            with rem.tunnel(
                lport, remote_socket, dhost="localhost", reverse=True
            ) as tun:
                assert tun.lport == lport
                if not dynamic_dport:
                    assert tun.dport == remote_socket
                else:
                    assert_is_port(tun.dport)
                assert tun.reverse

                remote_send_af_inet = f"""import socket
s = socket.socket()
s.connect(("localhost", {tun.dport}))
s.send("{message}".encode("ascii"))
s.close()
"""
                (rem.python["-u"] << remote_send_af_inet).popen()
                tunnel_server.join(timeout=1)
                assert queue.get() == message

    def test_get(self):
        with self._connect() as rem:
            assert str(rem["ls"]) == str(rem.get("ls"))
            assert str(rem["ls"]) == str(rem.get("not_a_valid_process_234", "ls"))
            assert "ls" in rem
            assert "not_a_valid_process_234" not in rem

    def test_list_processes(self):
        with self._connect() as rem:
            assert list(rem.list_processes())

    def test_pgrep(self):
        with self._connect() as rem:
            assert list(rem.pgrep("ssh"))

    def test_nohup(self):
        with self._connect() as rem:
            sleep = rem["sleep"]
            sleep["5.793817"] & NOHUP(stdout=None, append=False)
            time.sleep(0.5)
            print(rem["ps"]("aux"))
            assert list(rem.pgrep("5.793817"))
            time.sleep(6)
            assert not list(rem.pgrep("5.793817"))

    def test_bound_env(self):
        with self._connect() as rem:
            printenv = rem["printenv"]
            with rem.env(FOO="hello"):
                assert printenv.with_env(BAR="world")("FOO") == "hello\n"
                assert printenv.with_env(BAR="world")("BAR") == "world\n"
                assert printenv.with_env(FOO="sea", BAR="world")("FOO") == "sea\n"
                assert printenv.with_env(FOO="sea", BAR="world")("BAR") == "world\n"

            assert rem.cmd.pwd.with_cwd("/")() == "/\n"
            assert rem.cmd.pwd["-L"].with_env(A="X").with_cwd("/")() == "/\n"

    @pytest.mark.skipif(
        "useradd" not in local, reason="System does not have useradd (Mac?)"
    )
    def test_sshpass(self):
        with local.as_root():
            local["useradd"]("-m", "-b", "/tmp", "testuser")

        try:
            with local.as_root():
                try:
                    (local["passwd"] << "123456")("--stdin", "testuser")
                except ProcessExecutionError:
                    # some versions of passwd don't support --stdin, nothing to do in this case
                    logging.warning("passwd failed")
                    return

            with SshMachine("localhost", user="testuser", password="123456") as rem:
                assert rem["pwd"]().strip() == "/tmp/testuser"
        finally:
            with local.as_root():
                local["userdel"]("-r", "testuser")


@skip_on_windows
class TestParamikoMachine(BaseRemoteMachineTest):
    def _connect(self):
        if paramiko is None:
            pytest.skip("System does not have paramiko installed")
        return ParamikoMachine(TEST_HOST, missing_host_policy=paramiko.AutoAddPolicy())

    def test_tunnel(self):
        with self._connect() as rem:
            p = rem.python["-c", self.TUNNEL_PROG_AF_INET].popen()
            try:
                port = int(p.stdout.readline().strip())
            except ValueError:
                print(p.communicate())
                raise

            s = rem.connect_sock(port)
            s.send(b"world")
            data = s.recv(100)
            s.close()

        print(p.communicate())
        assert data == b"hello world"

    def test_piping(self):
        with self._connect() as rem:
            try:
                rem["ls"] | rem["cat"]
            except (NotImplementedError, TypeError):
                pass
            else:
                pytest.fail("Should not pipe")

    @pytest.mark.xfail(message="Not working yet")
    def test_encoding(self):
        with self._connect() as rem:
            unicode_half = b"\xc2\xbd".decode("utf8")

            ret = rem["bash"]("-c", 'echo -e "\xc2\xbd"')
            assert ret == f"{unicode_half}\n"

            ret = list(rem["bash"]["-c", 'echo -e "\xc2\xbd"'].popen())
            assert ret == [[f"{unicode_half}\n", None]]

    def test_path_open_remote_write_local_read(self):
        with self._connect() as rem:
            with rem.tempdir() as remote_tmpdir, local.tempdir() as tmpdir:
                assert remote_tmpdir.is_dir()
                assert tmpdir.is_dir()
                data = b"hello world"
                with (remote_tmpdir / "bar.txt").open("wb") as f:
                    f.write(data)
                rem.download((remote_tmpdir / "bar.txt"), (tmpdir / "bar.txt"))
                assert (tmpdir / "bar.txt").open("rb").read() == data

            assert not remote_tmpdir.exists()
            assert not tmpdir.exists()

    def test_path_open_local_write_remote_read(self):
        with self._connect() as rem:
            with rem.tempdir() as remote_tmpdir, local.tempdir() as tmpdir:
                assert remote_tmpdir.is_dir()
                assert tmpdir.is_dir()
                data = b"hello world"
                with (tmpdir / "bar.txt").open("wb") as f:
                    f.write(data)
                rem.upload((tmpdir / "bar.txt"), (remote_tmpdir / "bar.txt"))
                assert (remote_tmpdir / "bar.txt").open("rb").read() == data

            assert not remote_tmpdir.exists()
            assert not tmpdir.exists()



from __future__ import annotations

import locale

# High performance method for English (no translation needed)
loc = locale.getlocale()[0]
if loc is None or loc.startswith("en") or loc == "C":

    class NullTranslation:
        def gettext(self, str1: str) -> str:  # pylint: disable=no-self-use
            return str1

        def ngettext(self, str1, strN, n):  # pylint: disable=no-self-use
            if n == 1:
                return str1.replace("{0}", str(n))

            return strN.replace("{0}", str(n))

    def get_translation_for(
        package_name: str,  # noqa: ARG001
    ) -> NullTranslation:
        return NullTranslation()

else:
    import gettext
    from importlib.resources import as_file, files

    def get_translation_for(package_name: str) -> gettext.NullTranslations:  # type: ignore[misc]
        """
        Find and return gettext translation for package
        """
        assert loc is not None

        if "." in package_name:
            package_name = ".".join(package_name.split(".")[:-1])

        localedir = None

        with as_file(files(package_name) / "i18n") as mydir:
            for localedir in mydir, None:
                localefile = gettext.find(package_name, localedir, languages=[loc])
                if localefile:
                    break

            return gettext.translation(
                package_name, localedir=localedir, languages=[loc], fallback=True
            )



from __future__ import annotations

import os
import time

import psutil
import pytest

from plumbum import NOHUP, local

try:
    from plumbum.cmd import bash, echo
except ImportError:
    bash = None
    echo = None
from plumbum._testtools import skip_on_windows
from plumbum.path.utils import delete


@skip_on_windows
class TestNohupLocal:
    def read_file(self, filename):
        assert filename in os.listdir(".")
        with open(filename) as f:
            return f.read()

    @pytest.mark.usefixtures("testdir")
    def test_slow(self):
        delete("nohup.out")
        sp = bash["slow_process.bash"]
        sp & NOHUP
        time.sleep(0.5)
        assert self.read_file("slow_process.out") == "Starting test\n1\n"
        assert self.read_file("nohup.out") == "1\n"
        time.sleep(1)
        assert self.read_file("slow_process.out") == "Starting test\n1\n2\n"
        assert self.read_file("nohup.out") == "1\n2\n"
        time.sleep(2)
        delete("nohup.out", "slow_process.out")

    def test_append(self):
        delete("nohup.out")
        output = echo["This is output"]
        output & NOHUP
        time.sleep(0.2)
        assert self.read_file("nohup.out") == "This is output\n"
        output & NOHUP
        time.sleep(0.2)
        assert self.read_file("nohup.out") == "This is output\n" * 2
        delete("nohup.out")

    def test_redir(self):
        delete("nohup_new.out")
        output = echo["This is output"]

        output & NOHUP(stdout="nohup_new.out")
        time.sleep(0.2)
        assert self.read_file("nohup_new.out") == "This is output\n"
        delete("nohup_new.out")

        (output > "nohup_new.out") & NOHUP
        time.sleep(0.2)
        assert self.read_file("nohup_new.out") == "This is output\n"
        delete("nohup_new.out")

        output & NOHUP
        time.sleep(0.2)
        assert self.read_file("nohup.out") == "This is output\n"
        delete("nohup.out")

    def test_closed_filehandles(self):
        proc = psutil.Process()
        file_handles_prior = proc.num_fds()
        sleep_proc = local["sleep"]["1"] & NOHUP
        sleep_proc.wait()
        file_handles_after = proc.num_fds()
        assert file_handles_prior >= file_handles_after



from __future__ import annotations

import os
import platform
import sys

LINUX = sys.platform.startswith("linux")
MACOS = sys.platform.startswith("darwin")
WIN = sys.platform.startswith("win32") or sys.platform.startswith("cygwin")

CPYTHON = platform.python_implementation() == "CPython"
PYPY = platform.python_implementation() == "PyPy"

IS_A_TTY = sys.stdin.isatty()
HAS_CHOWN = hasattr(os, "chown")



from __future__ import annotations

import errno
import glob
import logging
import os
import shutil
import urllib.parse as urlparse
import urllib.request as urllib
from contextlib import contextmanager

from plumbum.lib import IS_WIN32
from plumbum.path.base import FSUser, Path
from plumbum.path.remote import RemotePath

try:
    from grp import getgrgid, getgrnam
    from pwd import getpwnam, getpwuid
except ImportError:

    def getpwuid(_x):  # type: ignore[misc]
        return (None,)

    def getgrgid(_x):  # type: ignore[misc]
        return (None,)

    def getpwnam(_x):  # type: ignore[misc]
        raise OSError("`getpwnam` not supported")

    def getgrnam(_x):  # type: ignore[misc]
        raise OSError("`getgrnam` not supported")


logger = logging.getLogger("plumbum.local")

_EMPTY = object()


# ===================================================================================================
# Local Paths
# ===================================================================================================
class LocalPath(Path):
    """The class implementing local-machine paths"""

    CASE_SENSITIVE = not IS_WIN32

    def __new__(cls, *parts):
        if (
            len(parts) == 1
            and isinstance(parts[0], cls)
            and not isinstance(parts[0], LocalWorkdir)
        ):
            return parts[0]
        if not parts:
            raise TypeError("At least one path part is required (none given)")
        if any(isinstance(path, RemotePath) for path in parts):
            raise TypeError(f"LocalPath cannot be constructed from {parts!r}")
        return super().__new__(
            cls, os.path.normpath(os.path.join(*(str(p) for p in parts)))
        )

    @property
    def _path(self):
        return str(self)

    def _get_info(self):
        return self._path

    def _form(self, *parts):
        return LocalPath(*parts)

    @property
    def name(self):
        return os.path.basename(str(self))

    @property
    def dirname(self):
        return LocalPath(os.path.dirname(str(self)))

    @property
    def suffix(self):
        return os.path.splitext(str(self))[1]

    @property
    def suffixes(self):
        exts = []
        base = str(self)
        while True:
            base, ext = os.path.splitext(base)
            if ext:
                exts.append(ext)
            else:
                return list(reversed(exts))

    @property
    def uid(self):
        uid = self.stat().st_uid
        name = getpwuid(uid)[0]
        return FSUser(uid, name)

    @property
    def gid(self):
        gid = self.stat().st_gid
        name = getgrgid(gid)[0]
        return FSUser(gid, name)

    def join(self, *others):
        return LocalPath(self, *others)

    def list(self):
        return [self / fn for fn in os.listdir(str(self))]

    def iterdir(self):
        try:
            return (self / fn.name for fn in os.scandir(str(self)))
        except AttributeError:
            return (self / fn for fn in os.listdir(str(self)))

    def is_dir(self):
        return os.path.isdir(str(self))

    def is_file(self):
        return os.path.isfile(str(self))

    def is_symlink(self):
        return os.path.islink(str(self))

    def exists(self):
        return os.path.exists(str(self))

    def stat(self):
        return os.stat(str(self))

    def with_name(self, name):
        return LocalPath(self.dirname) / name

    @property
    def stem(self):
        return self.name.rsplit(os.path.extsep)[0]

    def with_suffix(self, suffix, depth=1):
        if (
            suffix and not suffix.startswith(os.path.extsep)
        ) or suffix == os.path.extsep:
            raise ValueError(f"Invalid suffix {suffix!r}")
        name = self.name
        depth = len(self.suffixes) if depth is None else min(depth, len(self.suffixes))
        for _ in range(depth):
            name, _ = os.path.splitext(name)
        return LocalPath(self.dirname) / (name + suffix)

    def glob(self, pattern):
        return self._glob(
            pattern,
            lambda pat: [
                LocalPath(m)
                for m in glob.glob(os.path.join(glob.escape(str(self)), pat))
            ],
        )

    def delete(self):
        if not self.exists():
            return
        if self.is_dir():
            shutil.rmtree(str(self))
        else:
            try:
                os.remove(str(self))
            except OSError as ex:  # pragma: no cover
                # file might already been removed (a race with other threads/processes)
                if ex.errno != errno.ENOENT:
                    raise

    def move(self, dst):
        if isinstance(dst, RemotePath):
            raise TypeError(f"Cannot move local path {self} to {dst!r}")
        shutil.move(str(self), str(dst))
        return LocalPath(dst)

    def copy(self, dst, override=None):
        if isinstance(dst, RemotePath):
            raise TypeError(f"Cannot copy local path {self} to {dst!r}")
        dst = LocalPath(dst)
        if override is False and dst.exists():
            raise TypeError("File exists and override was not specified")
        if override:
            dst.delete()
        if self.is_dir():
            shutil.copytree(str(self), str(dst))
        else:
            dst_dir = LocalPath(dst).dirname
            if not dst_dir.exists():
                dst_dir.mkdir()
            shutil.copy2(str(self), str(dst))
        return dst

    def mkdir(self, mode=0o777, parents=True, exist_ok=True):
        if not self.exists() or not exist_ok:
            try:
                if parents:
                    os.makedirs(str(self), mode)
                else:
                    os.mkdir(str(self), mode)
            except OSError as ex:  # pragma: no cover
                # directory might already exist (a race with other threads/processes)
                if ex.errno != errno.EEXIST or not exist_ok:
                    raise

    def open(self, mode="r", encoding=None):
        return open(
            str(self),
            mode,
            encoding=encoding,
        )

    def read(self, encoding=None, mode="r"):
        if encoding and "b" not in mode:
            mode = mode + "b"
        with self.open(mode) as f:
            data = f.read()
            if encoding:
                return data.decode(encoding)
            return data

    def write(self, data, encoding=None, mode=None):
        if encoding:
            data = data.encode(encoding)
        if mode is None:
            mode = "w" if isinstance(data, str) else "wb"
        with self.open(mode) as f:
            f.write(data)

    def touch(self):
        with open(str(self), "a", encoding="utf-8"):
            os.utime(str(self), None)

    def chown(self, owner=None, group=None, recursive=None):
        if not hasattr(os, "chown"):
            raise OSError("os.chown() not supported")
        uid = (
            self.uid
            if owner is None
            else (owner if isinstance(owner, int) else getpwnam(owner)[2])
        )
        gid = (
            self.gid
            if group is None
            else (group if isinstance(group, int) else getgrnam(group)[2])
        )
        os.chown(str(self), uid, gid)
        if recursive or (recursive is None and self.is_dir()):
            for subpath in self.walk():
                os.chown(str(subpath), uid, gid)

    def chmod(self, mode):
        if not hasattr(os, "chmod"):
            raise OSError("os.chmod() not supported")
        os.chmod(str(self), mode)

    def access(self, mode=0):
        return os.access(str(self), self._access_mode_to_flags(mode))

    def link(self, dst):
        if isinstance(dst, RemotePath):
            raise TypeError(
                f"Cannot create a hardlink from local path {self} to {dst!r}"
            )
        if hasattr(os, "link"):
            os.link(str(self), str(dst))
        else:
            from plumbum.machines.local import local

            # windows: use mklink
            if self.is_dir():
                local["cmd"]("/C", "mklink", "/D", "/H", str(dst), str(self))
            else:
                local["cmd"]("/C", "mklink", "/H", str(dst), str(self))

    def symlink(self, dst):
        if isinstance(dst, RemotePath):
            raise TypeError(
                f"Cannot create a symlink from local path {self} to {dst!r}"
            )
        if hasattr(os, "symlink"):
            os.symlink(str(self), str(dst))
        else:
            from plumbum.machines.local import local

            # windows: use mklink
            if self.is_dir():
                local["cmd"]("/C", "mklink", "/D", str(dst), str(self))
            else:
                local["cmd"]("/C", "mklink", str(dst), str(self))

    def unlink(self):
        try:
            if hasattr(os, "symlink") or not self.is_dir():
                os.unlink(str(self))
            else:
                # windows: use rmdir for directories and directory symlinks
                os.rmdir(str(self))
        except OSError as ex:  # pragma: no cover
            # file might already been removed (a race with other threads/processes)
            if ex.errno != errno.ENOENT:
                raise

    def as_uri(self, scheme="file"):
        return urlparse.urljoin(f"{scheme}://", urllib.pathname2url(str(self)))

    @property
    def drive(self):
        return os.path.splitdrive(str(self))[0]

    @property
    def root(self):
        return os.path.sep


class LocalWorkdir(LocalPath):
    """Working directory manipulator"""

    def __hash__(self):
        raise TypeError("unhashable type")

    def __new__(cls):
        return super().__new__(cls, os.getcwd())

    def chdir(self, newdir):
        """Changes the current working directory to the given one

        :param newdir: The destination director (a string or a ``LocalPath``)
        """
        if isinstance(newdir, RemotePath):
            raise TypeError(f"newdir cannot be {newdir!r}")
        logger.debug("Chdir to %s", newdir)
        os.chdir(str(newdir))
        return self.__class__()

    def getpath(self):
        """Returns the current working directory as a ``LocalPath`` object"""
        return LocalPath(self._path)

    @contextmanager
    def __call__(self, newdir):
        """A context manager used to ``chdir`` into a directory and then ``chdir`` back to
        the previous location; much like ``pushd``/``popd``.

        :param newdir: The destination directory (a string or a ``LocalPath``)
        """
        prev = self._path
        newdir = self.chdir(newdir)
        try:
            yield newdir
        finally:
            self.chdir(prev)
